import unittest

import numpy as np
from pydrake.all import MathematicalProgram, Meshcat, Solve

import manipulation.meshcat_utils as dut


class TestMeshcatUtils(unittest.TestCase):
    def test_plot_mathematical_program(self):
        prog = MathematicalProgram()
        x = prog.NewContinuousVariables(2)
        prog.AddCost(x.dot(x))
        prog.AddBoundingBoxConstraint(-2, 2, x)
        result = Solve(prog)

        meshcat = Meshcat()
        X, Y = np.meshgrid(np.linspace(-3, 3, 35), np.linspace(-3, 3, 31))
        # Test that plotting doesn't raise any exceptions
        dut.plot_mathematical_program(meshcat, "test", prog, X, Y, result)


if __name__ == "__main__":
    unittest.main()



import hashlib
import unittest

import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestSurvey(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(1.0)
    def test_dynamics(self):
        """Test Survey Code"""
        survey_code = self.notebook_locals["survey_code"]
        m = hashlib.sha1(survey_code.encode("utf-8"))
        self.assertEqual(
            m.hexdigest(),
            "415133a1dd0d559a2fbe766054892d0a6b16fee9",
            "wrong survey code!",
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestRRT(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(5)
    @timeout_decorator.timeout(60.0)
    def test_rrt(self):
        """Test RRT Planner"""
        # construct a new problem
        self.notebook_locals["IKSolver"]
        IiwaProblem = self.notebook_locals["IiwaProblem"]
        env = self.notebook_locals["env"]

        np.random.seed(0)
        q_init = env.q0
        q_goal = self.notebook_locals["q_goal"]

        gripper_setpoint = 0.1
        left_door_angle = -np.pi / 2
        right_door_angle = np.pi / 2

        iiwa_problem_test = IiwaProblem(
            q_start=q_init,
            q_goal=q_goal,
            gripper_setpoint=gripper_setpoint,
            left_door_angle=left_door_angle,
            right_door_angle=right_door_angle,
            is_visualizing=False,
        )
        print("Constructed problem object")
        # load student RRT method
        student_rrt = self.notebook_locals["rrt_planning"]
        student_path = None
        print("Started to solve the problem using RRT planner ...")
        for i in range(3):  # allow 3 attempts
            student_path = student_rrt(iiwa_problem_test, 500, 0.5)
            if student_path is not None:
                print("run 1 found a solution")
                break
            print("run {} fails to find a solution".format(i))
        self.assertTrue(
            expr=(student_path is not None),
            msg="computed path is None. This may imply that "
            "the RRT planner failse to find a solution "
            "out of 3 attempts.",
        )
        self.assertTrue(
            isinstance(student_path[-1], tuple),
            msg="please use tuple to store configurations."
            "The computed path should be a list of tuples",
        )
        self.assertTrue(
            student_path[-1] == iiwa_problem_test.goal,
            msg="the last configuration of the path " "does not equal to goal.",
        )
        self.assertTrue(
            student_path[0] == iiwa_problem_test.start,
            msg="the first configuration of the path "
            "does not match start configuration",
        )



import unittest

import timeout_decorator
import torch
from gradescope_utils.autograder_utils.decorators import weight
from torch.nn.functional import l1_loss


class TestVPG(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(2)
    @timeout_decorator.timeout(60.0)
    def test_policy_loss(self):
        """Testing the policy loss"""
        PolicyEstimator = self.notebook_locals["PolicyEstimator"]
        student_policy_loss = self.notebook_locals["util_compute_policy_loss"]
        obs = torch.ones(3, 2) * 2.0
        actions = torch.ones(3, 1) * 3.0
        returns = torch.arange(3, 6).float()
        policy_f = PolicyEstimator(num_hidden=1, hidden_dim=2, obs_dim=2, action_dim=1)
        # initialize to constant values
        for p in policy_f.parameters():
            torch.nn.init.ones_(p)
        loss_student = student_policy_loss(policy_f, obs, actions, returns)
        self.assertTrue(
            torch.abs(loss_student - 7.6758) < 0.1,
            "The policy loss computation is incorrect",
        )

    @weight(2)
    @timeout_decorator.timeout(60.0)
    def test_value_loss(self):
        """Testing the value loss"""
        ValueEstimator = self.notebook_locals["ValueEstimator"]
        value_f = ValueEstimator(num_hidden=1, hidden_dim=2, obs_dim=2, action_dim=1)
        student_value_loss = self.notebook_locals["util_compute_value_loss"]
        # initialize to constant values
        for p in value_f.parameters():
            torch.nn.init.ones_(p)

        # test tensors
        obs = torch.ones(3, 2) * 2.0
        returns = torch.arange(3, 6).float()

        loss_student = student_value_loss(value_f, obs, returns)
        self.assertTrue(
            torch.abs(loss_student - 2.0317) < 0.1,
            "The value loss computation is incorrect",
        )

    @weight(4)
    @timeout_decorator.timeout(60.0)
    def test_advantage_function(self):
        """Testing advantage function"""
        compute_advantages = self.notebook_locals["compute_advantages"]
        discount = 0.99
        gae_lambda = 1.0
        max_episode_length = 3
        baselines = torch.FloatTensor([[1, 2, 3], [2, 1, 1]])
        rewards = torch.FloatTensor([[9, 8, 5], [6, 6, 6]])
        student_sol = compute_advantages(
            discount, gae_lambda, max_episode_length, baselines, rewards
        )
        reference_sol = torch.FloatTensor(
            [[20.8205, 10.9500, 2.9900], [15.8206, 10.9400, 5.9900]]
        )

        l1_error = l1_loss(input=student_sol[:, :-1], target=reference_sol[:, :-1])
        print(l1_error)
        self.assertLess(
            l1_error,
            5.0,
            "computation of the advantage " "function is incorrect",
        )



import dataclasses as dc
import os
import sys
import typing
import warnings
from copy import copy
from functools import partial

import numpy as np
from drake import (
    lcmt_iiwa_command,
    lcmt_iiwa_status,
    lcmt_image_array,
    lcmt_schunk_wsg_command,
    lcmt_schunk_wsg_status,
)
from pydrake.all import (
    ApplyCameraConfig,
    ApplyLcmBusConfig,
    ApplyMultibodyPlantConfig,
    ApplyVisualizationConfig,
    BaseField,
    CameraConfig,
    CameraInfo,
    DepthImageToPointCloud,
    Diagram,
    DiagramBuilder,
    DrakeLcmParams,
    FlattenModelDirectives,
    Gain,
    GetScopedFrameByName,
    IiwaCommandSender,
    IiwaControlMode,
    IiwaDriver,
    IiwaStatusReceiver,
    InverseDynamicsController,
    Joint,
    LcmBuses,
    LcmImageArrayToImages,
    LcmPublisherSystem,
    LcmSubscriberSystem,
    LeafSystem,
    MakeMultibodyStateToWsgStateSystem,
    Meshcat,
    MeshcatPointCloudVisualizer,
    ModelDirective,
    ModelDirectives,
    ModelInstanceIndex,
    ModelInstanceInfo,
    MultibodyPlant,
    MultibodyPlantConfig,
    MultibodyPositionToGeometryPose,
    Multiplexer,
    OutputPort,
    ParseIiwaControlMode,
    Parser,
    PdControllerGains,
    ProcessModelDirectives,
    RigidTransform,
    RobotDiagram,
    RobotDiagramBuilder,
    SceneGraph,
    SchunkWsgCommandSender,
    SchunkWsgDriver,
    SchunkWsgPositionController,
    SchunkWsgStatusReceiver,
    SharedPointerSystem,
    SimIiwaDriver,
    SimulatorConfig,
    VisualizationConfig,
    WeldJoint,
    ZeroForceDriver,
    position_enabled,
    torque_enabled,
)
from pydrake.common.yaml import yaml_load_typed

from manipulation.directives_tree import DirectivesTree
from manipulation.systems import ExtractPose
from manipulation.utils import ConfigureParser


@dc.dataclass
class JointPidControllerGains:
    """Defines the Proportional-Integral-Derivative gains for a single joint.

    Args:
        kp: The proportional gain.
        ki: The integral gain.
        kd: The derivative gain.
    """

    kp: float = 100  # Position gain
    ki: float = 1  # Integral gain
    kd: float = 20  # Velocity gain


@dc.dataclass
class InverseDynamicsDriver:
    """A simulation-only driver that adds the InverseDynamicsController to the
    station and exports the output ports. Multiple model instances can be driven with a
    single controller using `instance_name1+instance_name2` as the key; the output ports
    will be named similarly."""

    # Must have one element for every (named) actuator in the model_instance.
    gains: typing.Mapping[str, JointPidControllerGains] = dc.field(default_factory=dict)


@dc.dataclass
class JointPdControllerGains:
    """Defines the Proportional-Derivative gains for a single joint.

    Args:
        kp: The proportional gain.
        kd: The derivative gain.
    """

    kp: float = 0  # Position gain
    kd: float = 0  # Velocity gain


@dc.dataclass
class JointStiffnessDriver:
    """A simulation-only driver that sets up MultibodyPlant to act as if it is
    being controlled with a JointStiffnessController. The MultibodyPlant must
    be using SAP as the (discrete-time) contact solver.

    Args:
        gains: A mapping of {actuator_name: JointPdControllerGains} for each
            actuator that should be controlled.
        hand_model_name: If set, then the gravity compensation will be turned
            off for this model instance (e.g. for a hand).
    """

    # Must have one element for every (named) actuator in the model_instance.
    gains: typing.Mapping[str, JointPdControllerGains] = dc.field(default_factory=dict)

    hand_model_name: str = ""


@dc.dataclass
class Scenario:
    """Defines the YAML format for a (possibly stochastic) scenario to be
    simulated.

    Args:
        random_seed: Random seed for any random elements in the scenario. The
            seed is always deterministic in the `Scenario`; a caller who wants
            randomness must populate this value from their own randomness.
        simulation_duration: The maximum simulation time (in seconds).  The
            simulator will attempt to run until this time and then terminate.
        simulator_config: Simulator configuration (integrator and publisher
            parameters).
        plant_config: Plant configuration (time step and contact parameters).
        directives: All of the fully deterministic elements of the simulation.
        lcm_buses: A map of {bus_name: lcm_params} for LCM transceivers to be
            used by drivers, sensors, etc.
        model_drivers: For actuated models, specifies where each model's
            actuation inputs come from, keyed on the ModelInstance name.
        cameras: Cameras to add to the scene (and broadcast over LCM). The key
            for each camera is a helpful mnemonic, but does not serve a
            technical role. The CameraConfig::name field is still the name that
            will appear in the Diagram artifacts.
        visualization: Visualization configuration.
    """

    random_seed: int = 0

    simulation_duration: float = np.inf

    simulator_config: SimulatorConfig = SimulatorConfig(
        max_step_size=0.01,
        use_error_control=False,
        accuracy=1.0e-2,
    )

    plant_config: MultibodyPlantConfig = MultibodyPlantConfig()

    directives: typing.List[ModelDirective] = dc.field(default_factory=list)

    # Opt-out of LCM by default.
    lcm_buses: typing.Mapping[str, DrakeLcmParams] = dc.field(
        default_factory=lambda: dict(default=DrakeLcmParams(lcm_url="memq://null"))
    )

    model_drivers: typing.Mapping[
        str,
        typing.Union[
            IiwaDriver,
            InverseDynamicsDriver,
            JointStiffnessDriver,
            SchunkWsgDriver,
            ZeroForceDriver,
        ],
    ] = dc.field(default_factory=dict)

    cameras: typing.Mapping[str, CameraConfig] = dc.field(default_factory=dict)

    camera_ids: typing.Mapping[str, str] = dc.field(default_factory=dict)

    visualization: VisualizationConfig = VisualizationConfig()


@dc.dataclass
class Directives:
    """Defines the YAML format for directives to be applied to a `Scenario`.

    Args:
        directives: A list of `ModelDirective` objects.
    """

    directives: typing.List[ModelDirective] = dc.field(default_factory=list)


def load_scenario(
    *,
    filename: str | None = None,
    data: str | None = None,
    scenario_name: str | None = None,
    defaults: Scenario = Scenario(),
):
    warnings.warn("load_scenario is deprecated. Use LoadScenario instead.")
    return LoadScenario(
        filename=filename, data=data, scenario_name=scenario_name, defaults=defaults
    )


def add_scenario(
    *,
    filename: str | None = None,
    data: str | None = None,
    scenario_name: str | None = None,
    defaults: Scenario = Scenario(),
):
    warnings.warn("add_directives is deprecated. Use AppendDirectives instead.")
    return AppendDirectives(
        filename=filename, data=data, scenario_name=scenario_name, defaults=defaults
    )


# TODO(russt): load from url (using packagemap).
def LoadScenario(
    *,
    filename: str | None = None,
    data: str | None = None,
    scenario_name: str | None = None,
    defaults: Scenario = Scenario(),
) -> Scenario:
    """Implements the command-line handling logic for scenario data.

    Args:
        filename: A yaml filename to load the scenario from.

        data: A yaml _string_ to load the scenario from. If both filename and string are
            specified, then the filename is parsed first, and then the string is _also_
            parsed, potentially overwriting defaults from the filename. Note: this will
            not append additional `directives`, it will replace them; see
            AppendDirectives.

        scenario_name: The name of the scenario/child to load from the yaml file. If
            None, then the entire file is loaded.

        defaults: A `Scenario` object to use as the default values.

    Returns:
        A `Scenario` object loaded from the given input arguments.
    """
    result = defaults
    if filename:
        result = yaml_load_typed(
            schema=Scenario,
            filename=filename,
            child_name=scenario_name,
            defaults=result,
            retain_map_defaults=True,
        )
    if data:
        result = yaml_load_typed(
            schema=Scenario,
            data=data,
            child_name=scenario_name,
            defaults=result,
            retain_map_defaults=True,
        )
    return result


def AppendDirectives(
    scenario: Scenario,
    *,
    filename: str | None = None,
    data: str | None = None,
    scenario_name: str | None = None,
) -> Scenario:
    """Append additional directives to an existing scenario.

    Args:
        scenario: The scenario to append to.

        filename: A yaml filename to load the directives from.

        data: A yaml string to load the directives from. If both filename and string are
            specified, then the filename is parsed first, and then the string is _also_
            parsed, presumably overwriting any directives from the filename.

        scenario_name: The name of the scenario/child to load from the yaml file. If
            None, then the entire file is loaded.

    Returns:
        The scenario with the additional directives appended.
    """
    d = Directives()
    if filename:
        d = yaml_load_typed(
            schema=Directives,
            filename=filename,
            child_name=scenario_name,
        )
    if data:
        d = yaml_load_typed(
            schema=Directives,
            data=data,
            child_name=scenario_name,
        )
    scenario.directives.extend(d.directives)
    return scenario


def _FreezeChildren(
    plant: MultibodyPlant,
    children_to_freeze: typing.List[str],
) -> None:
    """
    Freeze the joints of the given children in the plant.

    Freezing really means removing a joint belonging to any of the children,
    and replacing it with a weld joint.
    """
    if len(children_to_freeze) == 0:
        return

    # Enumerate joints that need to be frozen i.e. removed and replaced by weld
    # joints. These are joints of child instances that are not already welds.
    joints_to_freeze: typing.Set[Joint] = set()
    for child_instance_name in children_to_freeze:
        child_instance = plant.GetModelInstanceByName(child_instance_name)
        for joint_index in plant.GetJointIndices(child_instance):
            joint = plant.get_joint(joint_index)
            if joint.type_name() != "weld":
                joints_to_freeze.add(joint)

    # Before removing joints, we need to remove associated actuators.
    for actuator_index in plant.GetJointActuatorIndices():
        actuator = plant.get_joint_actuator(actuator_index)
        if actuator.joint() in joints_to_freeze:
            plant.RemoveJointActuator(actuator)

    # Remove non-weld joints and replace them with weld joints.
    for joint in joints_to_freeze:
        weld = WeldJoint(
            joint.name(),
            joint.frame_on_parent(),
            joint.frame_on_child(),
            RigidTransform(),
        )
        plant.RemoveJoint(joint)
        plant.AddJoint(weld)


def _PopulatePlantOrDiagram(
    plant: MultibodyPlant,
    parser: Parser,
    scenario: Scenario,
    model_instance_names: typing.List[str] | None,
    add_frozen_child_instances: bool = True,
    package_xmls: typing.List[str] = [],
    parser_preload_callback: typing.Callable[[Parser], None] | None = None,
    parser_prefinalize_callback: typing.Callable[[Parser], None] | None = None,
) -> None:
    """See MakeMultibodyPlant and MakeRobotDiagram for details."""
    if model_instance_names is None:
        assert not add_frozen_child_instances, (
            "add_frozen_child_instances cannot be used without specifying "
            "model_instance_names."
        )

    ApplyMultibodyPlantConfig(scenario.plant_config, plant)
    for p in package_xmls:
        parser.package_map().AddPackageXml(p)
    ConfigureParser(parser)
    if parser_preload_callback:
        parser_preload_callback(parser)

    # Make the plant for the iiwa controller to use.
    flattened_directives = FlattenModelDirectives(
        ModelDirectives(directives=scenario.directives), parser.package_map()
    ).directives

    if not model_instance_names is None:
        tree = DirectivesTree(flattened_directives)
        directives = tree.GetDirectivesFromModelsToRoot(model_instance_names)
        children_to_freeze = set()

        if add_frozen_child_instances:
            children_to_freeze, additional_directives = (
                tree.GetWeldedDescendantsAndDirectives(model_instance_names)
            )
            directives.update(additional_directives)

        sorted_directives = tree.TopologicallySortDirectives(directives)
    else:
        # Add all model instances.
        sorted_directives = flattened_directives
        children_to_freeze = set()

    ProcessModelDirectives(
        directives=ModelDirectives(directives=sorted_directives),
        parser=parser,
    )

    _FreezeChildren(plant, children_to_freeze)

    if parser_prefinalize_callback:
        parser_prefinalize_callback(parser)

    plant.Finalize()


def MakeMultibodyPlant(
    scenario: Scenario,
    *,
    model_instance_names: typing.List[str] | None = None,
    add_frozen_child_instances: bool = False,
    package_xmls: typing.List[str] = [],
    parser_preload_callback: typing.Callable[[Parser], None] | None = None,
    parser_prefinalize_callback: typing.Callable[[Parser], None] | None = None,
) -> MultibodyPlant:
    """Use a scenario to create a MultibodyPlant. This is intended, e.g., to facilitate
    easily building subsets of a scenario, for instance, to make a plant for a
    controller.

    Args:
        scenario: A Scenario structure, populated using the `load_scenario`
            method.

        model_instance_names: If specified, then only the named model instances
            will be added to the plant. Otherwise, all model instances will be added.
            `add_weld` directives connecting added model instances to each other or to
            the world are also preserved.

        add_frozen_child_instances: If True and model_instance_names is not None, then
            model_instances that are not listed in model_instance_names, but are welded
            to a model_instance that is listed, will be added to the plant; with all
            joints replaced by welded joints.

        package_xmls: A list of package.xml file paths that will be passed to
            the parser, using Parser.AddPackageXml().

        parser_preload_callback: A callback function that will be called after
            the Parser is created, but before any directives are processed. This can be
            used to add additional packages to the parser, or to add additional model
            directives.

        parser_prefinalize_callback: A callback function that will be called
            after the directives are processed, but before the plant is finalized. This
            can be used to add additional model directives.

    Returns:
        A MultibodyPlant populated from (a subset of) the scenario.
    """
    plant = MultibodyPlant(time_step=scenario.plant_config.time_step)
    parser = Parser(plant)
    _PopulatePlantOrDiagram(
        plant,
        parser,
        scenario,
        model_instance_names,
        add_frozen_child_instances,
        package_xmls,
        parser_preload_callback,
        parser_prefinalize_callback,
    )
    return plant


def MakeRobotDiagram(
    scenario: Scenario,
    *,
    model_instance_names: typing.List[str] | None = None,
    add_frozen_child_instances: bool = True,
    package_xmls: typing.List[str] = [],
    parser_preload_callback: typing.Callable[[Parser], None] | None = None,
    parser_prefinalize_callback: typing.Callable[[Parser], None] | None = None,
) -> RobotDiagram:
    """Use a scenario to create a RobotDiagram (MultibodyPlant + SceneGraph). This is
    intended, e.g., to facilitate easily building subsets of a scenario, for instance,
    to make a plant for a controller which needs to make collision queries.

    Args:
        scenario: A Scenario structure, populated using the `load_scenario`
            method.

        model_instance_names: If specified, then only the named model instances
            will be added to the plant. Otherwise, all model instances will be added.

        add_frozen_child_instances: If True and model_instance_names is not None, then
            model_instances that are not listed in model_instance_names, but are welded
            to a model_instance that is listed, will be added to the plant; with all
            joints replaced by welded joints.

        package_xmls: A list of package.xml file paths that will be passed to
            the parser, using Parser.AddPackageXml().

        parser_preload_callback: A callback function that will be called after
            the Parser is created, but before any directives are processed. This can be
            used to add additional packages to the parser, or to add additional model
            directives.

        parser_prefinalize_callback: A callback function that will be called
            after the directives are processed, but before the plant is finalized. This
            can be used to add additional model directives.

    Returns:
        A RobotDiagram populated from (a subset of) the scenario.
    """
    robot_builder = RobotDiagramBuilder(time_step=scenario.plant_config.time_step)
    plant = robot_builder().builder().plant()
    parser = robot_builder().parser()
    _PopulatePlantOrDiagram(
        plant,
        parser,
        scenario,
        model_instance_names,
        add_frozen_child_instances,
        package_xmls,
        parser_preload_callback,
        parser_prefinalize_callback,
    )
    return robot_builder.Build()


class _MultiplexState(LeafSystem):
    def __init__(self, plant: MultibodyPlant, model_instance_names: typing.List[str]):
        LeafSystem.__init__(self)
        self._total_states = 0
        self._model_instance_names = model_instance_names
        self._plant = plant
        for name in self._model_instance_names:
            model_instance = plant.GetModelInstanceByName(name)
            num_states = plant.num_multibody_states(model_instance)
            # The logic below assumes num_positions == num_velocities (though
            # it would be simple to generalize).
            assert plant.num_positions(model_instance) == plant.num_velocities(
                model_instance
            )
            self.DeclareVectorInputPort(name + ".state", num_states)
            self._total_states += num_states
        self.DeclareVectorOutputPort(
            "combined_state",
            self._total_states,
            self.CalcOutput,
        )

    def CalcOutput(self, context, output):
        # The order should should be [q, q, ..., v, v, ...].
        positions = np.zeros((self._total_states // 2,))
        velocities = np.zeros((self._total_states // 2,))
        for i, model_instance_name in enumerate(self._model_instance_names):
            model_instance_index = self._plant.GetModelInstanceByName(
                model_instance_name
            )
            state = self.get_input_port(i).Eval(context)
            num_q = len(state) // 2
            self._plant.SetPositionsInArray(
                model_instance_index, state[:num_q], positions
            )
            self._plant.SetVelocitiesInArray(
                model_instance_index, state[num_q:], velocities
            )
        output.SetFromVector(np.concatenate((positions, velocities)))


class _DemultiplexActuation(LeafSystem):
    def __init__(self, plant: MultibodyPlant, model_instance_names: typing.List[str]):
        LeafSystem.__init__(self)
        total_inputs = 0
        self._plant = plant
        for name in model_instance_names:
            model_instance_index = plant.GetModelInstanceByName(name)
            num_actuators = plant.num_actuators(model_instance_index)
            self.DeclareVectorOutputPort(
                name + ".input",
                num_actuators,
                partial(self.CalcOutput, model_instance_index=model_instance_index),
            )
            total_inputs += num_actuators
        self.DeclareVectorInputPort("combined_input", total_inputs)

    def CalcOutput(self, context, output, model_instance_index):
        input = self.get_input_port().Eval(context)
        output.SetFromVector(
            self._plant.GetActuationFromArray(model_instance_index, input)
        )


# TODO(russt): Use the c++ version pending https://github.com/RobotLocomotion/drake/issues/20055
def _ApplyDriverConfigSim(
    driver_config,  # See Scenario.model_drivers for typing
    model_instance_name: str,
    sim_plant: MultibodyPlant,
    scenario: Scenario,
    package_xmls: typing.List[str],
    builder: DiagramBuilder,
) -> None:
    if isinstance(driver_config, IiwaDriver):
        model_instance = sim_plant.GetModelInstanceByName(model_instance_name)
        num_iiwa_positions = sim_plant.num_positions(model_instance)

        # Make the plant for the iiwa controller to use.
        controller_plant = MakeMultibodyPlant(
            scenario=scenario,
            model_instance_names=[model_instance_name],
            add_frozen_child_instances=True,
            package_xmls=package_xmls,
        )
        # Keep the controller plant alive during the Diagram lifespan.
        builder.AddNamedSystem(
            f"{model_instance_name}_controller_plant_pointer_system",
            SharedPointerSystem(controller_plant),
        )

        control_mode = ParseIiwaControlMode(driver_config.control_mode)
        sim_iiwa_driver = SimIiwaDriver.AddToBuilder(
            plant=sim_plant,
            iiwa_instance=model_instance,
            controller_plant=controller_plant,
            builder=builder,
            ext_joint_filter_tau=0.01,
            desired_iiwa_kp_gains=np.full(num_iiwa_positions, 100),
            control_mode=control_mode,
        )
        for i in range(sim_iiwa_driver.num_input_ports()):
            port = sim_iiwa_driver.get_input_port(i)
            if not builder.IsConnectedOrExported(port):
                builder.ExportInput(port, f"{model_instance_name}.{port.get_name()}")
        for i in range(sim_iiwa_driver.num_output_ports()):
            port = sim_iiwa_driver.get_output_port(i)
            builder.ExportOutput(port, f"{model_instance_name}.{port.get_name()}")

    elif isinstance(driver_config, SchunkWsgDriver):
        model_instance = sim_plant.GetModelInstanceByName(model_instance_name)
        # Wsg controller.
        wsg_controller = builder.AddSystem(SchunkWsgPositionController())
        wsg_controller.set_name(model_instance_name + ".controller")
        builder.Connect(
            wsg_controller.get_generalized_force_output_port(),
            sim_plant.get_actuation_input_port(model_instance),
        )
        builder.Connect(
            sim_plant.get_state_output_port(model_instance),
            wsg_controller.get_state_input_port(),
        )
        builder.ExportInput(
            wsg_controller.get_desired_position_input_port(),
            model_instance_name + ".position",
        )
        builder.ExportInput(
            wsg_controller.get_force_limit_input_port(),
            model_instance_name + ".force_limit",
        )
        wsg_mbp_state_to_wsg_state = builder.AddSystem(
            MakeMultibodyStateToWsgStateSystem()
        )
        builder.Connect(
            sim_plant.get_state_output_port(model_instance),
            wsg_mbp_state_to_wsg_state.get_input_port(),
        )
        builder.ExportOutput(
            wsg_mbp_state_to_wsg_state.get_output_port(),
            model_instance_name + ".state_measured",
        )
        builder.ExportOutput(
            wsg_controller.get_grip_force_output_port(),
            model_instance_name + ".force_measured",
        )

    elif isinstance(driver_config, InverseDynamicsDriver):
        model_instance_names = model_instance_name.split("+")
        model_instances = [
            sim_plant.GetModelInstanceByName(n) for n in model_instance_names
        ]

        # Make the controller plant.
        controller_plant = MakeMultibodyPlant(
            scenario=scenario,
            model_instance_names=model_instance_names,
            add_frozen_child_instances=True,
            package_xmls=package_xmls,
        )

        # Add the controller

        # When using multiple model instances, the model instance name must be prefixed.
        # The strings should take the form {model_instance_name}_{joint_actuator_name}, as
        # prescribed by MultiBodyPlant::GetActuatorNames().
        add_model_instance_prefix = len(model_instance_names) > 1
        actuator_names = controller_plant.GetActuatorNames(add_model_instance_prefix)

        # Check that all actuator names are valid.
        for actuator_name in driver_config.gains.keys():
            if actuator_name not in actuator_names:
                raise ValueError(
                    f"Actuator '{actuator_name}' not found. Valid names are: {actuator_names}"
                )

        # Get gains for each joint from the config. Use default gains if it doesn't exist in the config.
        default_gains = JointPidControllerGains()
        gains: typing.List[JointPidControllerGains] = []
        for actuator_name in actuator_names:
            joint_gains = driver_config.gains.get(actuator_name, default_gains)
            gains.append(joint_gains)

        controller = builder.AddSystem(
            InverseDynamicsController(
                controller_plant,
                kp=[joint_gains.kp for joint_gains in gains],
                ki=[joint_gains.ki for joint_gains in gains],
                kd=[joint_gains.kd for joint_gains in gains],
                has_reference_acceleration=False,
            )
        )
        controller.set_name(model_instance_name + ".controller")
        if len(model_instances) == 1:
            builder.Connect(
                sim_plant.get_state_output_port(model_instances[0]),
                controller.get_input_port_estimated_state(),
            )
            builder.Connect(
                controller.get_output_port_control(),
                sim_plant.get_actuation_input_port(model_instances[0]),
            )
            builder.ExportOutput(
                sim_plant.get_state_output_port(model_instances[0]),
                model_instance_name + ".state_estimated",
            )
        else:
            combined_state = builder.AddSystem(
                _MultiplexState(controller_plant, model_instance_names)
            )
            combined_state.set_name(model_instance_name + ".combined_state")
            combined_input = builder.AddSystem(
                _DemultiplexActuation(controller_plant, model_instance_names)
            )
            combined_input.set_name(model_instance_name + ".combined_input")
            for index, model_instance in enumerate(model_instances):
                builder.Connect(
                    sim_plant.get_state_output_port(model_instance),
                    combined_state.get_input_port(index),
                )
                builder.Connect(
                    combined_input.get_output_port(index),
                    sim_plant.get_actuation_input_port(model_instance),
                )
            builder.Connect(
                combined_state.get_output_port(),
                controller.get_input_port_estimated_state(),
            )
            builder.Connect(
                controller.get_output_port_control(), combined_input.get_input_port()
            )
            builder.ExportOutput(
                combined_state.get_output_port(),
                model_instance_name + ".state_estimated",
            )

        builder.ExportInput(
            controller.get_input_port_desired_state(),
            model_instance_name + ".desired_state",
        )

    elif isinstance(driver_config, JointStiffnessDriver):
        model_instance = sim_plant.GetModelInstanceByName(model_instance_name)

        # PD gains and gravity comp are set in ApplyPrefinalizeDriverConfigsSim

        builder.ExportInput(
            sim_plant.get_desired_state_input_port(model_instance),
            model_instance_name + ".desired_state",
        )
        builder.ExportInput(
            sim_plant.get_actuation_input_port(model_instance),
            model_instance_name + ".tau_feedforward",
        )
        builder.ExportOutput(
            sim_plant.get_state_output_port(model_instance),
            model_instance_name + ".state_estimated",
        )


def _ApplyDriverConfigsSim(
    *,
    driver_configs,  # See Scenario.model_drivers for typing
    sim_plant: MultibodyPlant,
    scenario: Scenario,
    package_xmls: typing.List[str],
    builder: DiagramBuilder,
) -> None:
    for model_instance_name, driver_config in driver_configs.items():
        _ApplyDriverConfigSim(
            driver_config=driver_config,
            model_instance_name=model_instance_name,
            sim_plant=sim_plant,
            scenario=scenario,
            package_xmls=package_xmls,
            builder=builder,
        )


def _ApplyPrefinalizeDriverConfigSim(
    driver_config,  # See Scenario.model_drivers for typing
    model_instance_name: str,
    sim_plant: MultibodyPlant,
    directives: typing.List[ModelDirective],
    models_from_directives_map: typing.Mapping[str, typing.List[ModelInstanceInfo]],
    package_xmls: typing.List[str],
    builder: DiagramBuilder,
) -> None:
    if isinstance(driver_config, JointStiffnessDriver):
        model_instance = sim_plant.GetModelInstanceByName(model_instance_name)

        # Set PD gains.
        for name, gains in driver_config.gains.items():
            actuator = sim_plant.GetJointActuatorByName(name, model_instance)
            actuator.set_controller_gains(PdControllerGains(p=gains.kp, d=gains.kd))

        # Turn off gravity to model (perfect) gravity compensation.
        sim_plant.set_gravity_enabled(model_instance, False)
        if driver_config.hand_model_name:
            sim_plant.set_gravity_enabled(
                sim_plant.GetModelInstanceByName(driver_config.hand_model_name),
                False,
            )


def _ApplyPrefinalizeDriverConfigsSim(
    *,
    driver_configs,  # See Scenario.model_drivers for typing
    sim_plant: MultibodyPlant,
    directives: typing.List[ModelDirective],
    models_from_directives: typing.Mapping[str, typing.List[ModelInstanceInfo]],
    package_xmls: typing.List[str],
    builder: DiagramBuilder,
) -> None:
    models_from_directives_map = dict(
        [(info.model_name, info) for info in models_from_directives]
    )
    for model_instance_name, driver_config in driver_configs.items():
        _ApplyPrefinalizeDriverConfigSim(
            driver_config,
            model_instance_name,
            sim_plant,
            directives,
            models_from_directives_map,
            package_xmls,
            builder,
        )


def _ApplyCameraConfigSim(
    *, config: CameraConfig, builder: DiagramBuilder, lcm_buses: LcmBuses
) -> None:
    # Always opt-out of lcm in this workflow.
    this_config = copy(config)
    this_config.lcm_bus = "opt_out"
    ApplyCameraConfig(config=this_config, builder=builder, lcm_buses=lcm_buses)

    camera_sys = builder.GetSubsystemByName(f"rgbd_sensor_{config.name}")
    builder.ExportOutput(
        camera_sys.color_image_output_port(), f"{config.name}.rgb_image"
    )
    builder.ExportOutput(
        camera_sys.depth_image_32F_output_port(), f"{config.name}.depth_image"
    )
    builder.ExportOutput(
        camera_sys.label_image_output_port(), f"{config.name}.label_image"
    )


def MakeHardwareStation(
    scenario: Scenario,
    meshcat: Meshcat | None = None,
    *,
    package_xmls: typing.List[str] = [],
    hardware: bool = False,
    parser_preload_callback: typing.Callable[[Parser], None] | None = None,
    parser_prefinalize_callback: typing.Callable[[Parser], None] | None = None,
    prebuild_callback: typing.Callable[[DiagramBuilder], None] | None = None,
) -> RobotDiagram:
    """Make a diagram encapsulating a simulation of (or the communications
    interface to/from) a physical robot, including sensors and controllers.

    Args:
        scenario: A Scenario structure, populated using the LoadScenario method.

        meshcat: If not None, then ApplyVisualizationConfig will called to add
            visualizers to the subdiagram using this meshcat instance.

        package_xmls: A list of package.xml file paths that will be passed to
            the parser, using Parser.AddPackageXml().

        hardware: If True, then the HardwareStationInterface diagram will be
            returned. Otherwise, the HardwareStation diagram will be returned.

        parser_preload_callback: A callback function that will be called after
            the Parser is created, but before any directives are processed. This
            can be used to add additional packages to the parser, or to add
            additional model directives.

        parser_prefinalize_callback: A callback function that will be called
            after the directives are processed, but before the plant is
            finalized. This can be used to add additional model directives.

        prebuild_callback: A callback function that will be called after the
            diagram builder is created, but before the diagram is built. This
            can be used to add additional systems to the diagram.

    Returns:
        If `hardware=False`, (the default) returns a HardwareStation diagram
        containing:
        - A MultibodyPlant with populated via the directives in `scenario`.
        - A SceneGraph
        - The default Drake visualizers
        - Any robot / sensors drivers specified in the YAML description.

        If `hardware=True`, returns a HardwareStationInterface diagram containing
        the network interfaces to communicate directly with the hardware drivers.
    """
    if hardware:
        return _MakeHardwareStationInterface(
            scenario=scenario, meshcat=meshcat, package_xmls=package_xmls
        )

    robot_builder = RobotDiagramBuilder(time_step=scenario.plant_config.time_step)
    builder = robot_builder.builder()
    sim_plant = robot_builder.plant()
    scene_graph = robot_builder.scene_graph()
    ApplyMultibodyPlantConfig(scenario.plant_config, sim_plant)

    parser = Parser(sim_plant)
    for p in package_xmls:
        parser.package_map().AddPackageXml(p)
    ConfigureParser(parser)
    if parser_preload_callback:
        parser_preload_callback(parser)

    # Add model directives.
    added_models = ProcessModelDirectives(
        directives=ModelDirectives(directives=scenario.directives),
        parser=parser,
    )

    if parser_prefinalize_callback:
        parser_prefinalize_callback(parser)

    # To support JointStiffnessControl, we need a pre-finalize version, too.
    _ApplyPrefinalizeDriverConfigsSim(
        driver_configs=scenario.model_drivers,
        sim_plant=sim_plant,
        directives=scenario.directives,
        models_from_directives=added_models,
        package_xmls=package_xmls,
        builder=builder,
    )

    sim_plant.Finalize()

    # For some Apply* functions in this workflow, we _never_ want LCM.
    scenario.lcm_buses["opt_out"] = DrakeLcmParams(lcm_url="memq://null")

    # Add LCM buses. (The simulator will handle polling the network for new
    # messages and dispatching them to the receivers, i.e., "pump" the bus.)
    lcm_buses = ApplyLcmBusConfig(lcm_buses=scenario.lcm_buses, builder=builder)

    # Add drivers.
    _ApplyDriverConfigsSim(
        driver_configs=scenario.model_drivers,
        sim_plant=sim_plant,
        scenario=scenario,
        package_xmls=package_xmls,
        builder=builder,
    )

    # Setup a virtual display if needed (for simulating cameras)
    if scenario.cameras and sys.platform == "linux" and os.getenv("DISPLAY") is None:
        from pyvirtualdisplay import Display

        virtual_display = Display(visible=0, size=(1400, 900))
        virtual_display.start()

    # Add scene cameras.
    for _, camera in scenario.cameras.items():
        _ApplyCameraConfigSim(config=camera, builder=builder, lcm_buses=lcm_buses)

    # Add visualization.
    if meshcat:
        ApplyVisualizationConfig(
            scenario.visualization, builder, meshcat=meshcat, lcm_buses=lcm_buses
        )

    # Export "cheat" ports.
    builder.ExportInput(
        sim_plant.get_applied_generalized_force_input_port(),
        "applied_generalized_force",
    )
    builder.ExportInput(
        sim_plant.get_applied_spatial_force_input_port(),
        "applied_spatial_force",
    )
    # Export any actuation (non-empty) input ports that are not already
    # connected (e.g. by a driver).
    for i in range(sim_plant.num_model_instances()):
        port = sim_plant.get_actuation_input_port(ModelInstanceIndex(i))
        if port.size() > 0 and not builder.IsConnectedOrExported(port):
            builder.ExportInput(port, port.get_name())
    # Export all MultibodyPlant output ports.
    for i in range(sim_plant.num_output_ports()):
        builder.ExportOutput(
            sim_plant.get_output_port(i),
            sim_plant.get_output_port(i).get_name(),
        )
    # Export the only SceneGraph output port.
    builder.ExportOutput(scene_graph.get_query_output_port(), "query_object")

    if prebuild_callback:
        prebuild_callback(builder)

    diagram = robot_builder.Build()
    diagram.set_name("station")
    return diagram


def NegatedPort(
    builder: DiagramBuilder, output_port: OutputPort, prefix: str = ""
) -> OutputPort:
    """Negates the output of the given port.

    Args:
        builder: The diagram builder to add the negation system to.
        output_port: The output port to negate.
        prefix: The prefix to use for the negation system.

    Returns:
        The output port of the negation system.
    """
    negater = builder.AddNamedSystem(
        f"sign_flip_{prefix}{output_port.get_name()}", Gain(-1, size=output_port.size())
    )
    builder.Connect(output_port, negater.get_input_port())
    return negater.get_output_port()


# TODO(russt): Use the c++ version pending https://github.com/RobotLocomotion/drake/issues/20055
def _ApplyDriverConfigInterface(
    driver_config,  # See Scenario.model_drivers for typing
    model_instance_name: str,
    lcm_buses: LcmBuses,
    builder: DiagramBuilder,
) -> None:
    if isinstance(driver_config, IiwaDriver):
        lcm = lcm_buses.Find("Driver for " + model_instance_name, driver_config.lcm_bus)

        # Publish IIWA command.
        # Note on publish period: IIWA driver won't respond faster than 1000Hz in
        # torque_only mode and 200Hz in other modes
        control_mode = ParseIiwaControlMode(driver_config.control_mode)
        publish_period = 0.005
        if control_mode == IiwaControlMode.kTorqueOnly:
            publish_period = 0.001
        iiwa_command_sender = builder.AddSystem(
            IiwaCommandSender(control_mode=control_mode)
        )
        iiwa_command_publisher = builder.AddSystem(
            LcmPublisherSystem.Make(
                channel="IIWA_COMMAND",
                lcm_type=lcmt_iiwa_command,
                lcm=lcm,
                publish_period=publish_period,
                use_cpp_serializer=True,
            )
        )
        iiwa_command_publisher.set_name(model_instance_name + ".command_publisher")
        if position_enabled(control_mode):
            builder.ExportInput(
                iiwa_command_sender.get_position_input_port(),
                model_instance_name + ".position",
            )
        if torque_enabled(control_mode):
            builder.ExportInput(
                iiwa_command_sender.get_torque_input_port(),
                model_instance_name + ".torque",
            )
        builder.Connect(
            iiwa_command_sender.get_output_port(),
            iiwa_command_publisher.get_input_port(),
        )
        # Receive IIWA status and populate the output ports.
        iiwa_status_receiver = builder.AddSystem(IiwaStatusReceiver())
        iiwa_status_receiver.set_name(model_instance_name + ".status_receiver")
        iiwa_status_subscriber = builder.AddSystem(
            LcmSubscriberSystem.Make(
                channel="IIWA_STATUS",
                lcm_type=lcmt_iiwa_status,
                lcm=lcm,
                use_cpp_serializer=True,
                wait_for_message_on_initialization_timeout=10,
            )
        )
        iiwa_status_subscriber.set_name(model_instance_name + ".status_subscriber")

        builder.ExportOutput(
            iiwa_status_receiver.get_position_commanded_output_port(),
            model_instance_name + ".position_commanded",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_position_measured_output_port(),
            model_instance_name + ".position_measured",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_velocity_estimated_output_port(),
            model_instance_name + ".velocity_estimated",
        )
        # These are *negative* w.r.t. the conventions outlined in
        # drake/manipulation/README.
        builder.ExportOutput(
            NegatedPort(
                builder=builder,
                output_port=iiwa_status_receiver.get_torque_commanded_output_port(),
                prefix=model_instance_name,
            ),
            model_instance_name + ".torque_commanded",
        )
        builder.ExportOutput(
            NegatedPort(
                builder=builder,
                output_port=iiwa_status_receiver.get_torque_measured_output_port(),
                prefix=model_instance_name,
            ),
            model_instance_name + ".torque_measured",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_torque_external_output_port(),
            model_instance_name + ".torque_external",
        )

        iiwa_state_mux: Multiplexer = builder.AddSystem(Multiplexer([7, 7]))
        builder.Connect(
            iiwa_status_receiver.get_position_measured_output_port(),
            iiwa_state_mux.get_input_port(0),
        )
        builder.Connect(
            iiwa_status_receiver.get_velocity_estimated_output_port(),
            iiwa_state_mux.get_input_port(1),
        )
        builder.ExportOutput(
            iiwa_state_mux.get_output_port(), model_instance_name + ".state_estimated"
        )

        builder.Connect(
            iiwa_status_subscriber.get_output_port(),
            iiwa_status_receiver.get_input_port(),
        )
    if isinstance(driver_config, SchunkWsgDriver):
        lcm = lcm_buses.Find("Driver for " + model_instance_name, driver_config.lcm_bus)

        # Publish WSG command.
        wsg_command_sender = builder.AddSystem(SchunkWsgCommandSender())
        wsg_command_publisher = builder.AddSystem(
            LcmPublisherSystem.Make(
                channel="SCHUNK_WSG_COMMAND",
                lcm_type=lcmt_schunk_wsg_command,
                lcm=lcm,
                publish_period=0.05,  # Schunk driver won't respond faster than 20Hz
                use_cpp_serializer=True,
            )
        )
        wsg_command_publisher.set_name(model_instance_name + ".command_publisher")
        builder.ExportInput(
            wsg_command_sender.get_position_input_port(),
            model_instance_name + ".position",
        )
        builder.ExportInput(
            wsg_command_sender.get_force_limit_input_port(),
            model_instance_name + ".force_limit",
        )
        builder.Connect(
            wsg_command_sender.get_output_port(0),
            wsg_command_publisher.get_input_port(),
        )

        # Receive WSG status and populate the output ports.
        wsg_status_receiver = builder.AddSystem(SchunkWsgStatusReceiver())
        wsg_status_receiver.set_name(model_instance_name + ".status_receiver")
        wsg_status_subscriber = builder.AddSystem(
            LcmSubscriberSystem.Make(
                channel="SCHUNK_WSG_STATUS",
                lcm_type=lcmt_schunk_wsg_status,
                lcm=lcm,
                use_cpp_serializer=True,
                wait_for_message_on_initialization_timeout=10,
            )
        )
        wsg_status_subscriber.set_name(model_instance_name + ".status_subscriber")
        builder.ExportOutput(
            wsg_status_receiver.get_state_output_port(),
            model_instance_name + ".state_measured",
        )
        builder.ExportOutput(
            wsg_status_receiver.get_force_output_port(),
            model_instance_name + ".force_measured",
        )
        builder.Connect(
            wsg_status_subscriber.get_output_port(),
            wsg_status_receiver.get_input_port(0),
        )


def _ApplyDriverConfigsInterface(
    *,
    driver_configs,  # See Scenario.model_drivers for typing
    lcm_buses: LcmBuses,
    builder: DiagramBuilder,
) -> None:
    for model_instance_name, driver_config in driver_configs.items():
        _ApplyDriverConfigInterface(
            driver_config,
            model_instance_name,
            lcm_buses,
            builder,
        )


class ConcatenateVectors(LeafSystem):
    """
    Concatenates multiple input vectors into a single output vector.
    """

    def __init__(self, input_vector_sizes: typing.List[int]):
        """
        Args:
            input_vector_sizes (List[int]): A list of input vector sizes. This
                system creates an input port corresponding to each size.
        """
        super().__init__()

        self.output_vector_size = sum(input_vector_sizes)

        # One input port per input vector size.
        for i, input_size in enumerate(input_vector_sizes):
            self.DeclareVectorInputPort(
                f"input_vector_{i + 1}",
                input_size,
            )

        # Single output port.
        self.DeclareVectorOutputPort(
            "concatenated_vector",
            self.output_vector_size,
            self.Calc,
        )

    def Calc(self, context, output):
        output_vector = np.zeros(self.output_vector_size)

        start_index = 0
        for i in range(self.num_input_ports()):
            input_vector = self.get_input_port(i).Eval(context)
            end_index = start_index + len(input_vector)
            output_vector[start_index:end_index] = input_vector
            start_index = end_index
        output.SetFromVector(output_vector)


def _WireDriverStatusReceiversToToPose(
    model_instance_names: typing.List[str],
    builder: DiagramBuilder,
    plant: MultibodyPlant,
    to_pose: MultibodyPositionToGeometryPose,
):

    # Create a ConcatenateVector system to concatenate positions for each model
    # instance.
    input_vector_sizes: typing.List[int] = []
    for model_instance_name in model_instance_names:
        model_instance_index = plant.GetModelInstanceByName(model_instance_name)
        num_model_positions = plant.num_positions(model_instance_index)
        input_vector_sizes.append(num_model_positions)
    concatenator = builder.AddSystem(ConcatenateVectors(input_vector_sizes))

    # Wire the status receiver for each model driver to the concatenator.
    for i, model_instance_name in enumerate(model_instance_names):
        status_receiver = builder.GetSubsystemByName(
            f"{model_instance_name}.status_receiver"
        )
        if isinstance(status_receiver, IiwaStatusReceiver):
            builder.Connect(
                status_receiver.get_position_measured_output_port(),
                concatenator.get_input_port(i),
            )
        elif isinstance(status_receiver, SchunkWsgStatusReceiver):
            builder.Connect(
                status_receiver.get_state_output_port(),
                concatenator.get_input_port(i),
            )
        else:
            raise ValueError(f"Unknown status subscriber: {status_receiver}")

    # Wire the concatenator to the to_pose system.
    builder.Connect(
        concatenator.get_output_port(),
        to_pose.get_input_port(),
    )


def _ApplyCameraLcmIdInterface(
    camera_config,  # See Scenario.cameras for typing
    camera_id: str,
    lcm_buses: LcmBuses,
    builder: DiagramBuilder,
) -> None:
    lcm = lcm_buses.Find("Driver for " + camera_config.name, camera_config.lcm_bus)

    camera_data_receiver = builder.AddSystem(LcmImageArrayToImages())
    camera_data_receiver.set_name(camera_config.name + ".data_receiver")
    camera_data_subscriber = builder.AddSystem(
        LcmSubscriberSystem.Make(
            channel=camera_id,
            lcm_type=lcmt_image_array,
            lcm=lcm,
            use_cpp_serializer=True,
            wait_for_message_on_initialization_timeout=10,
        )
    )
    camera_data_subscriber.set_name(camera_config.name + ".data_subscriber")

    builder.ExportOutput(
        camera_data_receiver.color_image_output_port(),
        f"{camera_config.name}.rgb_image",
    )
    builder.ExportOutput(
        camera_data_receiver.depth_image_output_port(),
        f"{camera_config.name}.depth_image",
    )
    builder.ExportOutput(
        camera_data_receiver.label_image_output_port(),
        f"{camera_config.name}.label_image",
    )

    builder.Connect(
        camera_data_subscriber.get_output_port(),
        camera_data_receiver.image_array_t_input_port(),
    )


def _MakeHardwareStationInterface(
    scenario: Scenario,
    meshcat: Meshcat | None = None,
    *,
    package_xmls: typing.List[str] = [],
) -> Diagram:
    builder = DiagramBuilder()

    # Visualize plant if Meshcat is provided.
    if meshcat is not None:
        scene_graph = builder.AddNamedSystem("scene_graph", SceneGraph())
        plant = MultibodyPlant(time_step=scenario.plant_config.time_step)
        ApplyMultibodyPlantConfig(scenario.plant_config, plant)
        plant.RegisterAsSourceForSceneGraph(scene_graph)
        parser = Parser(plant)
        for p in package_xmls:
            parser.package_map().AddPackageXml(p)
        ConfigureParser(parser)

        # Add model directives.
        added_models = ProcessModelDirectives(
            directives=ModelDirectives(directives=scenario.directives),
            parser=parser,
        )

        # Now the plant is complete.
        plant.Finalize()

        to_pose = builder.AddSystem(MultibodyPositionToGeometryPose(plant))
        builder.Connect(
            to_pose.get_output_port(),
            scene_graph.get_source_pose_port(plant.get_source_id()),
        )

        config = VisualizationConfig()
        config.publish_contacts = False
        config.publish_inertia = False
        ApplyVisualizationConfig(
            config,
            builder=builder,
            plant=plant,
            scene_graph=scene_graph,
            meshcat=meshcat,
        )

    # Add LCM buses. (The simulator will handle polling the network for new
    # messages and dispatching them to the receivers, i.e., "pump" the bus.)
    lcm_buses = ApplyLcmBusConfig(lcm_buses=scenario.lcm_buses, builder=builder)

    # Add drivers.
    _ApplyDriverConfigsInterface(
        driver_configs=scenario.model_drivers,
        lcm_buses=lcm_buses,
        builder=builder,
    )

    if meshcat is not None:
        _WireDriverStatusReceiversToToPose(
            model_instance_names=scenario.model_drivers.keys(),
            builder=builder,
            plant=plant,
            to_pose=to_pose,
        )

    # Add camera ids
    for camera_name, camera_id in scenario.camera_ids.items():
        _ApplyCameraLcmIdInterface(
            scenario.cameras[camera_name],
            camera_id,
            lcm_buses,
            builder,
        )

    diagram = builder.Build()
    diagram.set_name("HardwareStationInterface")
    return diagram


def AddPointClouds(
    *,
    scenario: Scenario,
    station: Diagram,
    builder: DiagramBuilder,
    poses_output_port: OutputPort | None = None,
    meshcat: Meshcat | None = None,
) -> typing.Mapping[str, DepthImageToPointCloud]:
    """
    Adds one DepthImageToPointCloud system to the `builder` for each camera in `scenario`, and connects it to the respective camera station output ports.

    Args:
        scenario: A Scenario structure, populated using the `LoadScenario` method.

        station: A HardwareStation system (e.g. from MakeHardwareStation) that has already been added to `builder`.

        builder: The DiagramBuilder containing `station` into which the new systems will be added.

        poses_output_port: (optional) HardwareStation will have a body_poses output port iff it was created with `hardware=False`. Alternatively, one could create a MultibodyPositionsToGeometryPoses system to consume the position measurements; this optional input can be used to support that workflow.

        meshcat: If not None, then a MeshcatPointCloudVisualizer will be added to the builder using this meshcat instance.

    Returns:
        A mapping from camera name to the DepthImageToPointCloud system.
    """
    to_point_cloud = dict()
    for _, config in scenario.cameras.items():
        if not config.depth:
            return

        plant = station.GetSubsystemByName("plant")
        # frame names in local variables:
        # P for parent frame, B for base frame, C for camera frame.

        # Extract the camera extrinsics from the config struct.
        P = (
            GetScopedFrameByName(plant, config.X_PB.base_frame)
            if config.X_PB.base_frame
            else plant.world_frame()
        )
        X_PC = config.X_PB.GetDeterministicValue()

        # convert mbp frame to geometry frame
        body = P.body()
        plant.GetBodyFrameIdIfExists(body.index())
        # assert body_frame_id.has_value()

        X_BP = P.GetFixedPoseInBodyFrame()
        X_BC = X_BP @ X_PC

        intrinsics = CameraInfo(
            config.width,
            config.height,
            config.focal_x(),
            config.focal_y(),
            config.principal_point()[0],
            config.principal_point()[1],
        )

        to_point_cloud[config.name] = builder.AddSystem(
            DepthImageToPointCloud(
                camera_info=intrinsics,
                fields=BaseField.kXYZs | BaseField.kRGBs,
            )
        )
        to_point_cloud[config.name].set_name(f"{config.name}.point_cloud")

        builder.Connect(
            station.GetOutputPort(f"{config.name}.depth_image"),
            to_point_cloud[config.name].depth_image_input_port(),
        )
        builder.Connect(
            station.GetOutputPort(f"{config.name}.rgb_image"),
            to_point_cloud[config.name].color_image_input_port(),
        )

        if poses_output_port is None:
            # Note: this is a cheat port; it will only work in single process
            # mode.
            poses_output_port = station.GetOutputPort("body_poses")

        camera_pose = builder.AddSystem(ExtractPose(int(body.index()), X_BC))
        camera_pose.set_name(f"{config.name}.pose")
        builder.Connect(
            poses_output_port,
            camera_pose.get_input_port(),
        )
        builder.Connect(
            camera_pose.get_output_port(),
            to_point_cloud[config.name].GetInputPort("camera_pose"),
        )

        if meshcat:
            # Send the point cloud to meshcat for visualization, too.
            point_cloud_visualizer = builder.AddSystem(
                MeshcatPointCloudVisualizer(meshcat, f"{config.name}.cloud")
            )
            builder.Connect(
                to_point_cloud[config.name].point_cloud_output_port(),
                point_cloud_visualizer.cloud_input_port(),
            )

    return to_point_cloud



import unittest
from copy import deepcopy

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


def ransac_solution(
    point_cloud, model_fit_func, rng, tolerance=1e-3, max_iterations=500
):
    """
    Args:
      point_cloud is (3, N) numpy array
      tolerance is a float
      rng is a random number generator
      max_iterations is a (small) integer
      model_fit_func: the function to fit the model (point clouds)

    Returns:
      (4,) numpy array
    """
    best_ic = 0  # inlier count
    best_model = None  # plane equation ((4,) array)
    N = point_cloud.shape[1]  # number of points

    sample_size = 3

    point_cloud_1 = np.ones((N, 4))
    point_cloud_1[:, :3] = point_cloud.T
    for i in range(max_iterations):
        s = point_cloud[:, rng.integers(0, N - 1, size=sample_size)]
        m = model_fit_func(s)
        abs_distances = np.abs(np.dot(m, point_cloud_1.T))  # 1 x N
        inliner_count = np.sum(abs_distances < tolerance)

        if inliner_count > best_ic:
            best_ic = inliner_count
            best_model = m

    return best_ic, best_model


class TestRANSAC(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals
        self.ransac_student = self.notebook_locals["ransac"]
        self.fit_plane = self.notebook_locals["fit_plane"]

    @weight(4)
    @timeout_decorator.timeout(2.0)
    def test_ransac(self):
        """Test ransac method"""
        simple_cloud = np.array(
            [
                [0, 1, 0],
                [0, 0, 0],
                [1, 0, 0],
                [1, 1, 0],
                [0.2, 0.5, 0],
                [2, 4, 1],
            ]
        ).T

        rng = np.random.default_rng(135)
        _, plane_student = self.ransac_student(
            simple_cloud, self.fit_plane, rng, 1e-5, 100
        )
        rng = np.random.default_rng(135)
        _, plane_solution = ransac_solution(
            simple_cloud, self.fit_plane, rng, 1e-5, 100
        )
        self.assertTrue(
            np.array_equal(plane_solution, plane_student),
            "ransac implementation incorrect",
        )

    @weight(2)
    @timeout_decorator.timeout(8.0)
    def test_outlier_removal(self):
        """Test outlier removal"""
        # check whether outlier is in student's answer
        remove_plane = self.notebook_locals["remove_plane"]
        bunny_w_plane = self.notebook_locals["bunny_w_plane"]
        bunny_w_plane_copy = deepcopy(bunny_w_plane)

        rng = np.random.default_rng(135)
        student_answer = remove_plane(bunny_w_plane, self.ransac_student, rng, tol=1e-4)

        outliers = []
        plane_equation = np.array([0.0, 0.0, 1.0, 0.0])
        dst = plane_equation[:3].dot(bunny_w_plane_copy) + plane_equation[3]
        for i in range(len(dst)):
            if abs(dst[i]) < 1e-6:
                outliers.append(bunny_w_plane_copy[:, i])
        outliers = np.asarray(outliers)
        for pnt in student_answer.T:
            self.assertTrue(
                not (pnt in outliers),
                "found unremoved points on the planar surface",
            )



import fnmatch
import os
import shutil
import tempfile
import unittest
from pathlib import Path

try:
    import trimesh

    from manipulation.create_sdf_from_mesh import create_sdf_from_mesh

    trimesh_available = True
except ImportError:
    trimesh_available = False
    print("trimesh not found.")
    print("Consider 'pip install trimesh'.")


@unittest.skipIf(not trimesh_available, "Requires trimesh dependency.")
class CreateSDFFromMeshTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._tmp_dir = tempfile.mkdtemp()
        cls._mesh_path = os.path.join(cls._tmp_dir, "box.obj")
        box = trimesh.primitives.Box([1.0, 1.0, 1.0])
        box.export(cls._mesh_path)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls._tmp_dir)

    def test_create_sdf_from_mesh(self):
        create_sdf_from_mesh(
            mesh_path=Path(self._mesh_path),
            mass=1.0,
            scale=1.5,
            is_compliant=True,
            hydroelastic_modulus=1e8,
            hunt_crossley_dissipation=None,
            mu_dynamic=1.0,
            mu_static=None,
            preview_with_trimesh=False,
        )
        self.assertTrue(
            os.path.exists(self._mesh_path.replace("obj", "sdf")),
            "SDFormat file does not exist",
        )
        mesh_pieces_dir = self._mesh_path.replace(".obj", "_parts")
        self.assertTrue(
            os.path.exists(mesh_pieces_dir), "Mesh pieces dir does not exist"
        )
        num_mesh_pieces = len(fnmatch.filter(os.listdir(mesh_pieces_dir), "*.*"))
        self.assertTrue(
            num_mesh_pieces == 1,
            f"An incorrect number of mesh pieces were created ({num_mesh_pieces} pieces)",
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


def gt_don_loss(f, img_a, img_b, u_a, u_b, match, margin=2.0):
    # in SWE it's not great to copy answer code here but we are just grading
    phi_a, phi_b = f(img_a[None])[0], f(img_b[None])[0]
    match = match[:, 0]
    h, w, d = phi_a.shape
    phi_a = phi_a.reshape(-1, d)[u_a[:, 0] * w + u_a[:, 1]]
    phi_b = phi_b.reshape(-1, d)[u_b[:, 0] * w + u_b[:, 1]]
    norm = np.linalg.norm(phi_a - phi_b, axis=1)
    loss_matches = np.mean(norm[match] ** 2)
    loss_nonmatches = np.mean(np.clip(margin - norm[~match], a_min=0, a_max=None) ** 2)
    return loss_matches, loss_nonmatches


def gt_don_predict(f, img_a, img_b, u_a):
    phi_a, phi_b = f(img_a[None])[0], f(img_b[None])[0]
    h, w, d = phi_a.shape
    phi_query = phi_a[u_a[0], u_a[1]]
    phi_keys = phi_b.reshape(-1, d)
    norm = np.linalg.norm(phi_keys - phi_query[None], axis=1)
    idx = np.argmin(norm)
    u_b = np.array([idx // w, idx % w])
    return u_b


def f_factory(rand_array):
    def dummy_f(x: np.ndarray):
        if x.ndim == 3:
            print(
                "f takes in image of shape (N, H, W, 3), "
                "did you forget to add batch dimension?"
            )
        phi = rand_array.copy()[None]
        phi = np.tile(phi, (len(x), 1, 1, 1))
        phi[..., 1:4] += x
        return phi

    return dummy_f


def get_random_indices(batch_size, h, w):
    return np.stack(
        [
            np.random.randint(h, size=(batch_size,)),
            np.random.randint(w, size=(batch_size,)),
        ],
        -1,
    )


class TestContrastive(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals
        np.random.seed(0)
        self._f = f_factory(np.random.rand(480, 640, 128))
        self._img_a = np.random.rand(480, 640, 3)
        self._img_b = np.random.rand(480, 640, 3)
        self._u_a = get_random_indices(4096, 480, 640)
        self._u_b = get_random_indices(4096, 480, 640)
        self._match = np.random.rand(4096, 1) > 0.4

    @weight(4)
    @timeout_decorator.timeout(10.0)
    def test_don_loss(self):
        """Testing don_loss"""
        expected_sol = gt_don_loss(
            self._f,
            self._img_a,
            self._img_b,
            self._u_a,
            self._u_b,
            self._match,
            margin=6.0,
        )
        sol = self.notebook_locals["don_loss"](
            self._f,
            self._img_a,
            self._img_b,
            self._u_a,
            self._u_b,
            self._match,
            margin=6.0,
        )

        self.assertTrue(
            np.abs(expected_sol[0] - sol[0]) < 1e-8,
            "Computed loss_matches is incorrect",
        )

        self.assertTrue(
            np.abs(expected_sol[1] - sol[1]) < 1e-8,
            "Computed loss_nonmatches is incorrect",
        )

    @weight(4)
    @timeout_decorator.timeout(10.0)
    def test_don_predict(self):
        """Testing don_predict"""
        expected_sol = gt_don_predict(self._f, self._img_a, self._img_b, self._u_a[0])
        sol = self.notebook_locals["don_predict"](
            self._f, self._img_a, self._img_b, self._u_a[0]
        )
        self.assertTrue(
            type(sol) == np.ndarray,
            "don_predict must return a np.ndarray as stated in docstring",
        )
        self.assertTrue(
            (expected_sol == sol).all(), "Predicted pixel position incorrect"
        )



"""
This file contains a number of helper utilities to set up our various
experiments with less code.
"""

import os
import sys
import warnings
from enum import Enum

import numpy as np
from pydrake.all import (
    AddCompliantHydroelasticProperties,
    AddContactMaterial,
    Adder,
    AddMultibodyPlantSceneGraph,
    BallRpyJoint,
    BaseField,
    Box,
    CameraInfo,
    Capsule,
    ClippingRange,
    CoulombFriction,
    Cylinder,
    Demultiplexer,
    DepthImageToPointCloud,
    DepthRange,
    DepthRenderCamera,
    DiagramBuilder,
    DifferentialInverseKinematicsIntegrator,
    Frame,
    GeometryInstance,
    InverseDynamicsController,
    MakeMultibodyStateToWsgStateSystem,
    MakePhongIllustrationProperties,
    MakeRenderEngineVtk,
    ModelInstanceIndex,
    MultibodyPlant,
    Parser,
    PassThrough,
    PrismaticJoint,
    ProximityProperties,
    RenderCameraCore,
    RenderEngineVtkParams,
    RevoluteJoint,
    Rgba,
    RgbdSensor,
    RigidTransform,
    RollPitchYaw,
    RotationMatrix,
    SchunkWsgPositionController,
    SpatialInertia,
    Sphere,
    StateInterpolatorWithDiscreteDerivative,
    UnitInertia,
)

from manipulation.systems import AddIiwaDifferentialIK as systems_AddIiwaDifferentialIK
from manipulation.systems import ExtractPose
from manipulation.utils import ConfigureParser

ycb = [
    "003_cracker_box.sdf",
    "004_sugar_box.sdf",
    "005_tomato_soup_can.sdf",
    "006_mustard_bottle.sdf",
    "009_gelatin_box.sdf",
    "010_potted_meat_can.sdf",
]


def ExtractBodyPose(output_port, body_index, X_BA=RigidTransform()) -> ExtractPose:
    warnings.warn(
        "ExtractBodyPose is deprecated. Use manipulation.systems.ExtractPose instead."
    )
    return ExtractPose(body_index, X_BA)


def AddIiwa(plant, collision_model="no_collision"):
    parser = Parser(plant)
    iiwa = parser.AddModelsFromUrl(
        f"package://drake_models/iiwa_description/sdf/iiwa7_{collision_model}.sdf"
    )[0]
    plant.WeldFrames(plant.world_frame(), plant.GetFrameByName("iiwa_link_0"))

    # Set default positions:
    q0 = [0.0, 0.1, 0, -1.2, 0, 1.6, 0]
    index = 0
    for joint_index in plant.GetJointIndices(iiwa):
        joint = plant.get_mutable_joint(joint_index)
        if isinstance(joint, RevoluteJoint):
            joint.set_default_angle(q0[index])
            index += 1

    return iiwa


def AddPlanarIiwa(plant):
    urdf = "package://drake_models/iiwa_description/urdf/planar_iiwa14_spheres_dense_elbow_collision.urdf"

    parser = Parser(plant)
    iiwa = parser.AddModelsFromUrl(urdf)[0]
    plant.WeldFrames(plant.world_frame(), plant.GetFrameByName("iiwa_link_0"))

    # Set default positions:
    q0 = [0.1, -1.2, 1.6]
    index = 0
    for joint_index in plant.GetJointIndices(iiwa):
        joint = plant.get_mutable_joint(joint_index)
        if isinstance(joint, RevoluteJoint):
            joint.set_default_angle(q0[index])
            index += 1

    return iiwa


def AddTwoLinkIiwa(plant, q0=[0.1, -1.2]):
    parser = Parser(plant)
    ConfigureParser(parser)
    iiwa = parser.AddModelsFromUrl("package://manipulation/two_link_iiwa14.urdf")[0]
    plant.WeldFrames(plant.world_frame(), plant.GetFrameByName("iiwa_link_0"))

    # Set default positions:
    index = 0
    for joint_index in plant.GetJointIndices(iiwa):
        joint = plant.get_mutable_joint(joint_index)
        if isinstance(joint, RevoluteJoint):
            joint.set_default_angle(q0[index])
            index += 1

    return iiwa


class WsgPositions(Enum):
    OPEN = 0.107
    CLOSED = 0.002


def AddWsg(plant, iiwa_model_instance, roll=np.pi / 2.0, welded=False, sphere=False):
    parser = Parser(plant)
    ConfigureParser(parser)
    if welded:
        if sphere:
            file = "package://manipulation/schunk_wsg_50_welded_fingers_sphere.sdf"
        else:
            file = "package://manipulation/schunk_wsg_50_welded_fingers.sdf"
    else:
        file = (
            "package://drake_models/wsg_50_description/sdf/schunk_wsg_50_with_tip.sdf"
        )

    directives = f"""
directives:
- add_model:
    name: gripper
    file: {file}
"""
    gripper = parser.AddModelsFromString(directives, ".dmd.yaml")[0]

    X_7G = RigidTransform(RollPitchYaw(np.pi / 2.0, 0, roll), [0, 0, 0.09])
    plant.WeldFrames(
        plant.GetFrameByName("iiwa_link_7", iiwa_model_instance),
        plant.GetFrameByName("body", gripper),
        X_7G,
    )
    return gripper


def AddFloatingXyzJoint(plant, frame, instance, actuators=False):
    inertia = UnitInertia.SolidSphere(1.0)
    x_body = plant.AddRigidBody(
        "x",
        instance,
        SpatialInertia(mass=0, p_PScm_E=[0.0, 0.0, 0.0], G_SP_E=inertia),
    )
    x_joint = plant.AddJoint(
        PrismaticJoint("x", plant.world_frame(), x_body.body_frame(), [1, 0, 0])
    )
    if actuators:
        plant.AddJointActuator("x", x_joint)
    y_body = plant.AddRigidBody(
        "y",
        instance,
        SpatialInertia(mass=0, p_PScm_E=[0.0, 0.0, 0.0], G_SP_E=inertia),
    )
    y_joint = plant.AddJoint(
        PrismaticJoint("y", x_body.body_frame(), y_body.body_frame(), [0, 1, 0])
    )
    if actuators:
        plant.AddJointActuator("y", y_joint)
    z_joint = plant.AddJoint(PrismaticJoint("z", y_body.body_frame(), frame, [0, 0, 1]))
    if actuators:
        plant.AddJointActuator("z", z_joint)


def AddFloatingRpyJoint(plant, frame, instance):
    inertia = UnitInertia.SolidSphere(1.0)
    z_body = plant.AddRigidBody(
        "z",
        instance,
        SpatialInertia(mass=0, p_PScm_E=[0.0, 0.0, 0.0], G_SP_E=inertia),
    )
    AddFloatingXyzJoint(plant, z_body.body_frame(), instance)
    plant.AddJoint(BallRpyJoint("ball", z_body.body_frame(), frame))


def AddShape(plant, shape, name, mass=1, mu=1, color=[0.5, 0.5, 0.9, 1.0]):
    instance = plant.AddModelInstance(name)
    # TODO: Add a method to UnitInertia that accepts a geometry shape (unless
    # that dependency is somehow gross) and does this.
    if isinstance(shape, Box):
        inertia = UnitInertia.SolidBox(shape.width(), shape.depth(), shape.height())
    elif isinstance(shape, Cylinder):
        inertia = UnitInertia.SolidCylinder(shape.radius(), shape.length(), [0, 0, 1])
    elif isinstance(shape, Sphere):
        inertia = UnitInertia.SolidSphere(shape.radius())
    elif isinstance(shape, Capsule):
        inertia = UnitInertia.SolidCapsule(shape.radius(), shape.length(), [0, 0, 1])
    else:
        raise RuntimeError(f"need to write the unit inertia for shapes of type {shape}")
    body = plant.AddRigidBody(
        name,
        instance,
        SpatialInertia(mass=mass, p_PScm_E=np.array([0.0, 0.0, 0.0]), G_SP_E=inertia),
    )
    if plant.geometry_source_is_registered():
        proximity_properties = ProximityProperties()
        AddContactMaterial(1e4, 1e7, CoulombFriction(mu, mu), proximity_properties)
        AddCompliantHydroelasticProperties(0.01, 1e8, proximity_properties)
        plant.RegisterCollisionGeometry(
            body, RigidTransform(), shape, name, proximity_properties
        )

        plant.RegisterVisualGeometry(body, RigidTransform(), shape, name, color)

    return instance


def AddCamera(builder, scene_graph, X_WC, depth_camera=None, renderer=None):
    warnings.warn("Please use AddRgbdSensor instead.", warnings.DeprecationWarning)
    return AddRgbdSensor(builder, scene_graph, X_WC, depth_camera, renderer)


def AddRgbdSensor(
    builder,
    scene_graph,
    X_PC,
    depth_camera=None,
    renderer=None,
    parent_frame_id=None,
):
    """
    Adds a RgbdSensor to to the scene_graph at (fixed) pose X_PC relative to
    the parent_frame.  If depth_camera is None, then a default camera info will
    be used.  If renderer is None, then we will assume the name 'my_renderer',
    and create a VTK renderer if a renderer of that name doesn't exist.  If
    parent_frame is None, then the world frame is used.
    """
    if sys.platform == "linux" and os.getenv("DISPLAY") is None:
        from pyvirtualdisplay import Display

        virtual_display = Display(visible=0, size=(1400, 900))
        virtual_display.start()

    if not renderer:
        renderer = "my_renderer"

    if not parent_frame_id:
        parent_frame_id = scene_graph.world_frame_id()

    if not scene_graph.HasRenderer(renderer):
        scene_graph.AddRenderer(renderer, MakeRenderEngineVtk(RenderEngineVtkParams()))

    if not depth_camera:
        depth_camera = DepthRenderCamera(
            RenderCameraCore(
                renderer,
                CameraInfo(width=640, height=480, fov_y=np.pi / 4.0),
                ClippingRange(near=0.1, far=10.0),
                RigidTransform(),
            ),
            DepthRange(0.1, 10.0),
        )

    rgbd = builder.AddSystem(
        RgbdSensor(
            parent_id=parent_frame_id,
            X_PB=X_PC,
            depth_camera=depth_camera,
            show_window=False,
        )
    )

    builder.Connect(scene_graph.get_query_output_port(), rgbd.query_object_input_port())

    return rgbd


def AddRgbdSensors(
    builder,
    plant,
    scene_graph,
    also_add_point_clouds=True,
    model_instance_prefix="camera",
    depth_camera=None,
    renderer=None,
):
    """
    Adds a RgbdSensor to the first body in the plant for every model instance
    with a name starting with model_instance_prefix.  If depth_camera is None,
    then a default camera info will be used.  If renderer is None, then we will
    assume the name 'my_renderer', and create a VTK renderer if a renderer of
    that name doesn't exist.
    """
    if sys.platform == "linux" and os.getenv("DISPLAY") is None:
        from pyvirtualdisplay import Display

        virtual_display = Display(visible=0, size=(1400, 900))
        virtual_display.start()

    if not renderer:
        renderer = "my_renderer"

    if not scene_graph.HasRenderer(renderer):
        scene_graph.AddRenderer(renderer, MakeRenderEngineVtk(RenderEngineVtkParams()))

    if not depth_camera:
        depth_camera = DepthRenderCamera(
            RenderCameraCore(
                renderer,
                CameraInfo(width=640, height=480, fov_y=np.pi / 4.0),
                ClippingRange(near=0.1, far=10.0),
                RigidTransform(),
            ),
            DepthRange(0.1, 10.0),
        )

    for index in range(plant.num_model_instances()):
        model_instance_index = ModelInstanceIndex(index)
        model_name = plant.GetModelInstanceName(model_instance_index)

        if model_name.startswith(model_instance_prefix):
            body_index = plant.GetBodyIndices(model_instance_index)[0]
            rgbd = builder.AddSystem(
                RgbdSensor(
                    parent_id=plant.GetBodyFrameIdOrThrow(body_index),
                    X_PB=RigidTransform(),
                    depth_camera=depth_camera,
                    show_window=False,
                )
            )
            rgbd.set_name(model_name)

            builder.Connect(
                scene_graph.get_query_output_port(),
                rgbd.query_object_input_port(),
            )

            # Export the camera outputs
            builder.ExportOutput(
                rgbd.color_image_output_port(), f"{model_name}_rgb_image"
            )
            builder.ExportOutput(
                rgbd.depth_image_32F_output_port(), f"{model_name}_depth_image"
            )
            builder.ExportOutput(
                rgbd.label_image_output_port(), f"{model_name}_label_image"
            )

            if also_add_point_clouds:
                # Add a system to convert the camera output into a point cloud
                to_point_cloud = builder.AddSystem(
                    DepthImageToPointCloud(
                        camera_info=rgbd.default_depth_render_camera()
                        .core()
                        .intrinsics(),
                        fields=BaseField.kXYZs | BaseField.kRGBs,
                    )
                )
                builder.Connect(
                    rgbd.depth_image_32F_output_port(),
                    to_point_cloud.depth_image_input_port(),
                )
                builder.Connect(
                    rgbd.color_image_output_port(),
                    to_point_cloud.color_image_input_port(),
                )

                camera_pose = builder.AddSystem(ExtractPose(body_index))
                builder.Connect(
                    plant.get_body_poses_output_port(),
                    camera_pose.get_input_port(),
                )
                builder.Connect(
                    camera_pose.get_output_port(),
                    to_point_cloud.GetInputPort("camera_pose"),
                )

                # Export the point cloud output.
                builder.ExportOutput(
                    to_point_cloud.point_cloud_output_port(),
                    f"{model_name}_point_cloud",
                )


def SetTransparency(scene_graph, alpha, source_id, geometry_ids=None):
    inspector = scene_graph.model_inspector()
    if not geometry_ids:
        geometry_ids = inspector.GetAllGeometryIds()

    for gid in geometry_ids:
        if not inspector.BelongsToSource(gid, source_id):
            continue
        props = inspector.GetIllustrationProperties(gid)
        if props is None or not props.HasProperty("phong", "diffuse"):
            continue
        c = props.GetProperty("phong", "diffuse")
        new_color = Rgba(c.r(), c.g(), c.b(), alpha)
        props.UpdateProperty("phong", "diffuse", new_color)


# TODO(russt): Use Rgba instead of vector color.
def SetColor(scene_graph, color, source_id, geometry_ids=None):
    inspector = scene_graph.model_inspector()
    if not geometry_ids:
        geometry_ids = inspector.GetAllGeometryIds()

    for gid in geometry_ids:
        if not inspector.BelongsToSource(gid, source_id):
            continue
        props = inspector.GetIllustrationProperties(gid)
        if props is None or not props.HasProperty("phong", "diffuse"):
            continue
        new_color = Rgba(color[0], color[1], color[2], color[3])
        props.UpdateProperty("phong", "diffuse", new_color)


def AddTriad(
    source_id,
    frame_id,
    scene_graph,
    length=0.25,
    radius=0.01,
    opacity=1.0,
    X_FT=RigidTransform(),
    name="frame",
):
    """
    Adds illustration geometry representing the coordinate frame, with the
    x-axis drawn in red, the y-axis in green and the z-axis in blue. The axes
    point in +x, +y and +z directions, respectively.

    Args:
      source_id: The source registered with SceneGraph.
      frame_id: A geometry::frame_id registered with scene_graph.
      scene_graph: The SceneGraph with which we will register the geometry.
      length: the length of each axis in meters.
      radius: the radius of each axis in meters.
      opacity: the opacity of the coordinate axes, between 0 and 1.
      X_FT: a RigidTransform from the triad frame T to the frame_id frame F
      name: the added geometry will have names name + " x-axis", etc.
    """
    # x-axis
    X_TG = RigidTransform(RotationMatrix.MakeYRotation(np.pi / 2), [length / 2.0, 0, 0])
    geom = GeometryInstance(
        X_FT.multiply(X_TG), Cylinder(radius, length), name + " x-axis"
    )
    geom.set_illustration_properties(
        MakePhongIllustrationProperties([1, 0, 0, opacity])
    )
    scene_graph.RegisterGeometry(source_id, frame_id, geom)

    # y-axis
    X_TG = RigidTransform(RotationMatrix.MakeXRotation(np.pi / 2), [0, length / 2.0, 0])
    geom = GeometryInstance(
        X_FT.multiply(X_TG), Cylinder(radius, length), name + " y-axis"
    )
    geom.set_illustration_properties(
        MakePhongIllustrationProperties([0, 1, 0, opacity])
    )
    scene_graph.RegisterGeometry(source_id, frame_id, geom)

    # z-axis
    X_TG = RigidTransform([0, 0, length / 2.0])
    geom = GeometryInstance(
        X_FT.multiply(X_TG), Cylinder(radius, length), name + " z-axis"
    )
    geom.set_illustration_properties(
        MakePhongIllustrationProperties([0, 0, 1, opacity])
    )
    scene_graph.RegisterGeometry(source_id, frame_id, geom)


def AddMultibodyTriad(frame, scene_graph, length=0.25, radius=0.01, opacity=1.0):
    plant = frame.GetParentPlant()
    AddTriad(
        plant.get_source_id(),
        plant.GetBodyFrameIdOrThrow(frame.body().index()),
        scene_graph,
        length,
        radius,
        opacity,
        frame.GetFixedPoseInBodyFrame(),
        name=frame.name(),
    )


def AddIiwaDifferentialIK(
    builder: DiagramBuilder, plant: MultibodyPlant, frame: Frame | None = None
) -> DifferentialInverseKinematicsIntegrator:
    warnings.warn(
        "Please use manipulation.systems.AddIiwaDifferentialIK instead.",
        DeprecationWarning,
    )
    if frame is None:
        frame = plant.GetFrameByName("body")
    return systems_AddIiwaDifferentialIK(builder, plant, frame)


# This method is effectively deprecated, and should not be used. See MakeHardwareStation
# instead.
def MakeManipulationStation(
    model_directives=None,
    filename=None,
    time_step=0.002,
    iiwa_prefix="iiwa",
    wsg_prefix="wsg",
    camera_prefix="camera",
    prefinalize_callback=None,
    package_xmls=[],
):
    """
    Creates a manipulation station system, which is a sub-diagram containing:
      - A MultibodyPlant with populated via the Parser from the
        `model_directives` argument AND the `filename` argument.
      - A SceneGraph
      - For each model instance starting with `iiwa_prefix`, we add an
        additional iiwa controller system
      - For each model instance starting with `wsg_prefix`, we add an
        additional schunk controller system
      - For each body starting with `camera_prefix`, we add a RgbdSensor

    Args:
        builder: a DiagramBuilder

        model_directives: a string containing any model directives to be parsed

        filename: a string containing the name of an sdf, urdf, mujoco xml, or
        model directives yaml file.

        time_step: the standard MultibodyPlant time step.

        iiwa_prefix: Any model instances starting with `iiwa_prefix` will get
        an inverse dynamics controller, etc attached

        wsg_prefix: Any model instance starting with `wsg_prefix` will get a
        schunk controller

        camera_prefix: Any bodies in the plant (created during the
        plant_setup_callback) starting with this prefix will get a camera
        attached.

        prefinalize_callback: A function, setup(plant), that will be called
        with the multibody plant before calling finalize.  This can be useful
        for e.g. adding additional bodies/models to the simulation.

        package_xmls: A list of filenames to be passed to
        PackageMap.AddPackageXml().  This is useful if you need to add more
        models to your path (e.g. from your current working directory).
    """
    warnings.warn(
        "MakeManipulationStation is deprecated. Use MakeHardwareStation instead.",
        DeprecationWarning,
    )

    builder = DiagramBuilder()

    # Add (only) the iiwa, WSG, and cameras to the scene.
    plant, scene_graph = AddMultibodyPlantSceneGraph(builder, time_step=time_step)
    parser = Parser(plant)
    for p in package_xmls:
        parser.package_map().AddPackageXml(p)
    ConfigureParser(parser)
    if model_directives:
        parser.AddModelsFromString(model_directives, ".dmd.yaml")
    if filename:
        parser.AddModelsFromUrl(filename)
    if prefinalize_callback:
        prefinalize_callback(plant)
    plant.Finalize()

    for i in range(plant.num_model_instances()):
        model_instance = ModelInstanceIndex(i)
        model_instance_name = plant.GetModelInstanceName(model_instance)

        if model_instance_name.startswith(iiwa_prefix):
            num_iiwa_positions = plant.num_positions(model_instance)

            # I need a PassThrough system so that I can export the input port.
            iiwa_position = builder.AddSystem(PassThrough(num_iiwa_positions))
            builder.ExportInput(
                iiwa_position.get_input_port(),
                model_instance_name + "_position",
            )
            builder.ExportOutput(
                iiwa_position.get_output_port(),
                model_instance_name + "_position_commanded",
            )

            # Export the iiwa "state" outputs.
            demux = builder.AddSystem(
                Demultiplexer(2 * num_iiwa_positions, num_iiwa_positions)
            )
            builder.Connect(
                plant.get_state_output_port(model_instance),
                demux.get_input_port(),
            )
            builder.ExportOutput(
                demux.get_output_port(0),
                model_instance_name + "_position_measured",
            )
            builder.ExportOutput(
                demux.get_output_port(1),
                model_instance_name + "_velocity_estimated",
            )
            builder.ExportOutput(
                plant.get_state_output_port(model_instance),
                model_instance_name + "_state_estimated",
            )

            # Make the plant for the iiwa controller to use.
            controller_plant = MultibodyPlant(time_step=time_step)
            # TODO: Add the correct IIWA model (introspected from MBP)
            if plant.num_positions(model_instance) == 3:
                controller_iiwa = AddPlanarIiwa(controller_plant)
            else:
                controller_iiwa = AddIiwa(controller_plant)
            AddWsg(controller_plant, controller_iiwa, welded=True)
            controller_plant.Finalize()

            # Add the iiwa controller
            iiwa_controller = builder.AddSystem(
                InverseDynamicsController(
                    controller_plant,
                    kp=[100] * num_iiwa_positions,
                    ki=[1] * num_iiwa_positions,
                    kd=[20] * num_iiwa_positions,
                    has_reference_acceleration=False,
                )
            )
            iiwa_controller.set_name(model_instance_name + "_controller")
            builder.Connect(
                plant.get_state_output_port(model_instance),
                iiwa_controller.get_input_port_estimated_state(),
            )

            # Add in the feed-forward torque
            adder = builder.AddSystem(Adder(2, num_iiwa_positions))
            builder.Connect(
                iiwa_controller.get_output_port_control(),
                adder.get_input_port(0),
            )
            # Use a PassThrough to make the port optional (it will provide zero
            # values if not connected).
            torque_passthrough = builder.AddSystem(
                PassThrough([0] * num_iiwa_positions)
            )
            builder.Connect(
                torque_passthrough.get_output_port(), adder.get_input_port(1)
            )
            builder.ExportInput(
                torque_passthrough.get_input_port(),
                model_instance_name + "_torque",
            )
            builder.Connect(
                adder.get_output_port(),
                plant.get_actuation_input_port(model_instance),
            )

            # Add discrete derivative to command velocities.
            desired_state_from_position = builder.AddSystem(
                StateInterpolatorWithDiscreteDerivative(
                    num_iiwa_positions,
                    time_step,
                    suppress_initial_transient=True,
                )
            )
            desired_state_from_position.set_name(
                model_instance_name + "_desired_state_from_position"
            )
            builder.Connect(
                desired_state_from_position.get_output_port(),
                iiwa_controller.get_input_port_desired_state(),
            )
            builder.Connect(
                iiwa_position.get_output_port(),
                desired_state_from_position.get_input_port(),
            )

            # Export commanded torques.
            builder.ExportOutput(
                adder.get_output_port(),
                model_instance_name + "_torque_commanded",
            )
            builder.ExportOutput(
                adder.get_output_port(),
                model_instance_name + "_torque_measured",
            )

            builder.ExportOutput(
                plant.get_generalized_contact_forces_output_port(model_instance),
                model_instance_name + "_torque_external",
            )

        elif model_instance_name.startswith(wsg_prefix):
            # Wsg controller.
            wsg_controller = builder.AddSystem(SchunkWsgPositionController())
            wsg_controller.set_name(model_instance_name + "_controller")
            builder.Connect(
                wsg_controller.get_generalized_force_output_port(),
                plant.get_actuation_input_port(model_instance),
            )
            builder.Connect(
                plant.get_state_output_port(model_instance),
                wsg_controller.get_state_input_port(),
            )
            builder.ExportInput(
                wsg_controller.get_desired_position_input_port(),
                model_instance_name + "_position",
            )
            builder.ExportInput(
                wsg_controller.get_force_limit_input_port(),
                model_instance_name + "_force_limit",
            )
            wsg_mbp_state_to_wsg_state = builder.AddSystem(
                MakeMultibodyStateToWsgStateSystem()
            )
            builder.Connect(
                plant.get_state_output_port(model_instance),
                wsg_mbp_state_to_wsg_state.get_input_port(),
            )
            builder.ExportOutput(
                wsg_mbp_state_to_wsg_state.get_output_port(),
                model_instance_name + "_state_measured",
            )
            builder.ExportOutput(
                wsg_controller.get_grip_force_output_port(),
                model_instance_name + "_force_measured",
            )

    # Cameras.
    AddRgbdSensors(builder, plant, scene_graph, model_instance_prefix=camera_prefix)

    # Export "cheat" ports.
    builder.ExportOutput(scene_graph.get_query_output_port(), "query_object")
    builder.ExportOutput(plant.get_contact_results_output_port(), "contact_results")
    builder.ExportOutput(plant.get_state_output_port(), "plant_continuous_state")
    builder.ExportOutput(plant.get_body_poses_output_port(), "body_poses")

    diagram = builder.Build()
    diagram.set_name("ManipulationStation")
    return diagram



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight

from manipulation.utils import FindDataResource


def chamfer_dist(pc_a, pc_b):
    """
    pc_a of Size(3, N)
    pc_b of Size(3, M)
    """
    diff = np.linalg.norm(pc_a[:, :, None] - pc_b[:, None, :], axis=0) ** 2
    dist = np.mean(np.min(diff, axis=1)) + np.mean(np.min(diff, axis=0))
    return dist


class TestSegmentationAndGrasp(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(10.0)
    def test_get_merged_masked_pcd(self):
        """Test find_antipodal_pts"""
        predictions = self.notebook_locals["predictions"]
        cameras = self.notebook_locals["cameras"]
        get_merged_masked_pcd = self.notebook_locals["get_merged_masked_pcd"]

        rgb_ims = [c.rgb_im for c in cameras]
        depth_ims = [c.depth_im for c in cameras]
        project_depth_to_pC_funcs = [c.project_depth_to_pC for c in cameras]
        X_WCs = [c.X_WC for c in cameras]

        pcd_eval = get_merged_masked_pcd(
            predictions, rgb_ims, depth_ims, project_depth_to_pC_funcs, X_WCs
        )
        pcd_pts_eval = np.asarray(pcd_eval.xyzs()[:])
        pcd_colors_eval = np.asarray(pcd_eval.rgbs()[:])
        num_points_eval = pcd_pts_eval.shape[1]

        data_target = np.load(FindDataResource("segmentation_and_grasp_soln.npz"))
        pcd_pts_target = data_target["points"]
        data_target["colors"]
        num_points_target = pcd_pts_target.shape[1]

        # Allow some deviation in the number of points
        self.assertLessEqual(
            np.linalg.norm(num_points_target - num_points_eval),
            200,
            "Wrong number of points returned.",
        )

        # Make sure the sizes match
        min_num_pts = min(num_points_eval, num_points_target)

        self.assertLessEqual(
            chamfer_dist(
                pcd_pts_target[:, :min_num_pts], pcd_pts_eval[:, :min_num_pts]
            ),
            5e-4,
            "Point cloud points are not close enough to the solution values.",
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestHardwareStationIO(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(5)
    @timeout_decorator.timeout(1.0)
    def test_dynamics(self):
        """Test get_velocity implementation"""
        f = self.notebook_locals["get_velocity"]
        station = self.notebook_locals["station"]
        station_context = self.notebook_locals["station_context"]
        plant = self.notebook_locals["plant"]
        plant_context = self.notebook_locals["plant_context"]
        iiwa = plant.GetModelInstanceByName("iiwa")

        np.random.seed(7)
        for i in range(10):
            test_vel = np.random.rand(7)  # draw 7 random numbers
            plant.SetVelocities(plant_context, iiwa, test_vel)
            eval_vel = f(station, station_context)
            self.assertLessEqual(
                np.linalg.norm(test_vel - eval_vel),
                1e-6,
                "get_velocity implementation is not correct!",
            )



import warnings

from pydrake.gym import *  # noqa

warnings.warn(
    "manipulation.drake_gym has moved to pydrake.gym. This shim will be removed after 2023-12-31.",
    DeprecationWarning,
    stacklevel=2,
)



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
from pydrake.all import RigidTransform, RollPitchYaw, RotationMatrix


def least_squares_transform(scene, model):
    X_BA = RigidTransform()
    mu_m = np.mean(model, axis=1)
    mu_s = np.mean(scene, axis=1)

    W = (scene.T - mu_s).T @ (model.T - mu_m)
    U, Sigma, Vh = np.linalg.svd(W)
    R_star = U.dot(Vh)

    if np.linalg.det(R_star) < 0:
        Vh[-1] *= -1
        R_star = U.dot(Vh)

    t_star = mu_s - R_star.dot(mu_m)

    X_BA.set_rotation(RotationMatrix(R_star))
    X_BA.set_translation(t_star)

    return X_BA


def generate_arbitrary_transform(seed):
    R_BA = RollPitchYaw(seed * 7.363, seed * 1.35, seed * 5.47)
    p_BA = np.mod([seed * 2.13, seed * 3.3, seed * 5.225], 0.1) - 0.05
    X_BA = RigidTransform(R_BA, p_BA)
    return X_BA


class TestICP(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

        self.model = self.notebook_locals["pointcloud_model"]

    @weight(3)
    @timeout_decorator.timeout(5.0)
    def test_least_square(self):
        """Test least square transform"""
        f = self.notebook_locals["least_squares_transform"]
        nearest_neighbors = self.notebook_locals["nearest_neighbors"]

        # Run a battery of tests deterministically, making sure that the
        # condition number of the data matrix (the singular values of W) are
        # not too small.
        for i in range(10):
            X_BA = generate_arbitrary_transform(i)
            self.scene = X_BA.multiply(self.model)
            distances, indices = nearest_neighbors(self.scene, self.model)

            X_BA_test = f(self.scene, self.model[:, indices])
            X_BA_true = least_squares_transform(self.scene, self.model[:, indices])

            # check answer
            result = X_BA_true.inverse().multiply(X_BA_test)

            self.assertTrue(
                np.allclose(result.GetAsMatrix4(), np.eye(4)),
                "least square transform is incorrect",
            )

        # Test improper matrix
        result = f(self.model, np.diag([1, 1, -1]) @ self.model)
        self.assertTrue(
            result.rotation().IsValid(),
            "Your method returned an improper rotation matrix",
        )

    @weight(3)
    @timeout_decorator.timeout(10.0)
    def test_icp(self):
        """Test icp implementation"""
        f = self.notebook_locals["icp"]
        nearest_neighbors = self.notebook_locals["nearest_neighbors"]

        # It should be sufficient to test only one for ICP.
        X_BA = generate_arbitrary_transform(7)
        self.scene = X_BA.multiply(self.model)
        X_BA_test, mean_error_test, num_iters_test = f(self.scene, self.model)

        distances, indices = nearest_neighbors(
            self.scene, X_BA_test.multiply(self.model)
        )

        mean_error = np.mean(distances)
        assert mean_error < 1.3e-2, "ICP test failed with mean error {}".format(
            mean_error
        )



# Written by Caelan Garrett, modified a bit by Tomas Lozano-Perez
# yapf: disable
from math import atan2, cos, sin, sqrt

# Some utilities used in the class definitions, notably Polygon,


# is exp near zero?
def within(exp, threshold=.001):
    return abs(exp) < threshold


# Test that a list of polygon points is convex.
def is_convex(
        verts):  # Must be increasing except for one pair on the circular list
    angles = [(verts[i] - verts[i - 1]).perpendicular().angle()
              for i in range(len(verts))]
    out_of_order = 0
    for i in range(len(angles)):
        if angles[i - 1] > angles[i]:
            out_of_order += 1
            if out_of_order > 1:
                return False
    return True


# Test that the angles of the normals to the edges are monotonically
# increasing.
def order_increasing(verts):
    angles = [(verts[i] - verts[i - 1]).perpendicular().angle()
              for i in range(len(verts))]
    for i in range(len(angles) - 1):
        if angles[i] > angles[i + 1]:
            return verts[i + 1:] + verts[:i + 1]
    return verts


# Linear merge of two ordered lists, based on <=.  This is used to
# merge lists of edges by angles in the cspace computation for two
# convex polygons (the so-called "star algorithm.
def merge(a, b):
    result = []
    p_a = 0
    p_b = 0
    for i in range(len(a) + len(b)):
        if p_b >= len(b) or (p_a < len(a) and a[p_a] <= b[p_b]):
            result.append(a[p_a])
            p_a += 1
        else:
            result.append(b[p_b])
            p_b += 1
    return result


# Point (x,y) class, also used for vector computations (caveat emptor)
class Point:

    def __init__(self, x, y):
        self.x = float(x)
        self.y = float(y)

    def __add__(self, other):  # vector rum
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):  # vector difference
        return Point(self.x - other.x, self.y - other.y)

    def length(self):  # length of vector
        return sqrt(self.x * self.x + self.y * self.y)

    def angle(self):  # the angle from origin to point
        return atan2(self.y, self.x)

    def normalize(self):  # make unit vector
        return (1.0 / self.length()) * self

    def distance(self, other):  # distance between two points
        return (other - self).length()

    def dot(self, other):  # dot product of two vectors
        return self.x * other.x + self.y * other.y

    def cross(self,
              other):  # Returns z component of cross (everything else is zero)
        return (a.x * b.y - a.y * b.x)

    def perpendicular(self):  # Perpendicular vector
        return Point(self.y, -self.x)

    def rotate(self, angle):  # Rotate point around the origin
        current_angle = self.angle()
        return self.length() * Point(cos(current_angle + angle),
                                     sin(current_angle + angle))

    def rotate_about(self, point, angle):  # rotate about another point
        return (self - point).rotate(angle) + point

    def __rmul__(self, scale):  # scale vector by a constant
        return Point(scale * self.x, scale * self.y)

    def __neg__(self):  # negate vector
        return Point(-self.x, -self.y)

    def __eq__(self, other):  # equality test
        return isinstance(other, Point) and \
            self.x == other.x and self.y == other.y

    def __ne__(self, other):  # not equal
        return not self == other

    def draw(self, window, color='black', radius=3):  # draw the point
        window.drawPoint(self.x, self.y, color, radius=radius)

    def __repr__(self):
        return 'Point: (' + str(self.x) + ', ' + str(self.y) + ')'

    def __hash__(self):
        return str(self).__hash__()

    __str__ = __repr__


# An SO(2) pose ((x,y), theta)
class Pose:

    def __init__(self, position, orientation):
        self.position = position
        self.orientation = float(orientation)

    def __eq__(self, other):  # strict equality
        return isinstance(other, Pose) and \
            self.position == other.position and \
            self.orientation == other.orientation

    def __ne__(self, other):  # inequality
        return not self == other

    def draw(self, window, color='black', radius=3):  # draw the pose as a point
        self.position.draw(window, color=color, radius=radius)

    def __repr__(self):
        return 'Pose: (' + str(self.position) + ', ' + str(
            self.orientation) + ')'

    def __hash__(self):
        return str(self).__hash__()

    __str__ = __repr__


# A Line (segment) from start to end.
class Line:

    def __init__(self, start, end):
        self.start = start
        self.end = end
        self.aabb = None
        self.leq = None

    def get_aabb(self):  # bounding box of line
        if self.aabb is None:
            min_x = min(self.start.x, self.end.x)
            min_y = min(self.start.y, self.end.y)
            max_x = max(self.start.x, self.end.x)
            max_y = max(self.start.y, self.end.y)
            self.aabb = AABB(Point(min_x, min_y), Point(max_x, max_y))
        return self.aabb

    def get_leq(self):  # get line equation n.x + d = 0
        if self.leq is None:
            n = self.unit_vector().perpendicular()
            d = -n.dot(self.start)
            self.leq = (n, d)
        return self.leq

    def length(self):  # length of line segment
        return self.start.distance(self.end)

    def unit_vector(self):  # unit vector from start to end
        return (1.0 / self.length()) * (self.end - self.start)

    def left(self):  # left end point
        if self.start.x <= self.end.x:
            return self.start
        else:
            return self.end

    def right(self):  # right end point
        if self.start.x > self.end.x:
            return self.start
        else:
            return self.end

    def is_left_turn(self, point):  # is the point left of the line?
        (n, d) = self.get_leq()
        # left means in the negative half-space
        return n.dot(point) + d < 0

    def distance(self, pt):  # perp distance of pt to line
        if self.start == self.end:  # if line is a point
            return self.start.distance(pt)
        else:  # a real line
            (n, d) = self.get_leq()
            return n.dot(pt) + d

    def intersect(self, other, edges=False):  # do lines intersect?
        # edges = True counts a point being on an edge as collision
        if not self.get_aabb().collides(other.get_aabb()):
            # bounding boxes do not overlap
            return False
        d_1 = (self.end - self.start)
        perp_1 = -d_1.perpendicular()
        d_2 = other.end - other.start
        diff = (self.start - other.start)
        num = diff.dot(perp_1)
        denom = d_2.dot(perp_1)
        if denom == 0:
            return False
        b = (1.0 * num) / denom
        intersect = other.start + b * d_2
        if d_1.x == 0:
            a = (intersect - self.start).y / d_1.y
        else:
            a = (intersect - self.start).x / d_1.x
        if edges:
            return a >= 0 and a <= 1 and b >= 0 and b <= 1
        else:
            return a > 0 and a < 1 and b > 0 and b < 1

    # Cut one line with another, return the (up to) two segments
    def cut(self, other):
        if not self.get_aabb().collides(other.get_aabb()):
            return None
        d_1 = (self.end - self.start)
        perp_1 = -d_1.perpendicular()
        d_2 = other.end - other.start
        diff = (self.start - other.start)
        num = diff.dot(perp_1)
        denom = d_2.dot(perp_1)
        if denom == 0:
            return None
        b = (1.0 * num) / denom
        intersect = other.start + b * d_2
        if d_1.x == 0:
            a = (intersect - self.start).y / d_1.y
        else:
            a = (intersect - self.start).x / d_1.x
        if a > 0 and a < 1 and b > 0 and b < 1 and \
                not within((intersect - self.start).length()) and \
                not within((intersect - self.end).length()):
            # Sometimes the cut is just the line + point
            return [Line(self.start, intersect), Line(intersect, self.end)]
        else:
            return None

    def equal_leq(self, other):  # equality of line equations
        return self.get_leq() == other.get_leq()

    def midpoint(self):  # middle point of segment
        return (self.start + self.end) / 2

    def draw(self, window, color="black", width=3):  # draw
        window.drawLineSeg(self.start.x,
                           self.start.y,
                           self.end.x,
                           self.end.y,
                           color=color,
                           width=width)

    def __eq__(self, other):  # equality (directed)
        return isinstance(other, Line) and \
            self.start == other.start and self.end == other.end

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return 'Line: (' + str(self.start) + ', ' + str(self.end) + ')'

    def __hash__(self):
        return str(self).__hash__()

    __str__ = __repr__


# Axis-Aligned Bounding Box
class AABB:

    def __init__(self, min, max):
        self.min = min
        self.max = max
        self.verts = None

    def get_verts(self):  # vertices of box
        if self.verts is None:
            self.verts = [
                self.min,
                Point(self.max.x, self.min.y), self.max,
                Point(self.min.x, self.max.y)
            ]
        return self.verts

    def get_lines(self):  # edge lines of box
        points = self.get_verts()
        return [Line(points[i - 1], points[i]) for i in range(len(points))]

    def translate(self, reference):  # translate the box
        return AABB(self.min + point, self.max + point)

    def union(self, other):  # the union of two boxes
        return AABB(
            Point(min(self.min.x, other.min.x), min(self.min.y, other.min.y)),
            Point(max(self.max.x, other.max.y), max(self.max.y, other.max.y)))

    def intersection(self, other):  # intersection of two boxes
        if self.collides(other):
            (xlo1, ylo1, zlo1), (xhi1, yhi1, zhi1) = bb1
            (xlo2, ylo2, zlo2), (xhi2, yhi2, zhi2) = bb2
            return AABB(
                Point(max(self.min.x, other.min.x), max(self.min.y,
                                                        other.min.y)),
                Point(min(self.max.x, other.max.x), min(self.max.y,
                                                        other.max.y)))
        else:
            return None

    def volume(self):  # volume of the box
        return (self.max.x - self.min.x) * (self.max.y - self.min.y)

    def collides(self, other, edges=False):  # do two boxes overlap?
        return not (self.min.x > other.max.x or self.max.x < other.min.x
                    or self.min.y > other.max.y or self.max.y < other.min.y)

    def contains(self,
                 other,
                 edges=False):  # is some entity contained in the box
        # edges = True counts a point being on an edge as containment
        if isinstance(other, Point):
            if edges:
                return self.min.x <= other.x and self.min.y <= other.y and \
                    self.max.x >= other.x and self.max.y >= other.y
            else:
                return self.min.x < other.x and self.min.y < other.y and \
                    self.max.x > other.x and self.max.y > other.y
        elif isinstance(other, Line):  # Definition of convex object
            return self.contains(other.start, edges=edges) and \
                self.contains(other.end, edges=edges)
        elif isinstance(other, AABB):
            if edges:
                return (self.min.x <= other.min.x
                        and self.min.y <= other.min.y
                        and self.max.x >= other.max.x
                        and self.max.y >= other.max.y)
            else:
                return (self.min.x < other.min.x
                        and self.min.y < other.min.y
                        and self.max.x > other.max.x
                        and self.max.y > other.max.y)
        elif isinstance(other, Polygon):
            return all([self.contains(vert, edges) for vert in other.points])
        elif isinstance(other, Object):
            return all([self.contains(poly, edges) for poly in other.polys])
        else:
            raise Exception

    def draw(self, window, color="black", width=3):  # draw box
        verts = self.get_verts()
        for i in range(len(verts)):
            Line(verts[i - 1], verts[i]).draw(window, color=color, width=width)

    def __repr__(self):
        return 'AABB: (' + str(self.min) + ', ' + str(self.max) + ')'

    def __hash__(self):
        return str(self).__hash__()

    __str__ = __repr__


# A convex polygon
class Polygon:

    def __init__(self, points):
        assert is_convex(points)
        self.points = order_increasing(points)
        self.leqs = None
        self.aabb = None
        self.radius = None

    def get_aabb(self):  # get bounding box
        if self.aabb is None:
            min_x = min([point.x for point in self.points])
            min_y = min([point.y for point in self.points])
            max_x = max([point.x for point in self.points])
            max_y = max([point.y for point in self.points])
            self.aabb = AABB(Point(min_x, min_y), Point(max_x, max_y))
        return self.aabb

    def get_leqs(self):  # line equations, in normal form
        if self.leqs is None:
            self.leqs = []
            for i in range(len(self.points)):
                diff = self.points[i] - self.points[i - 1]
                n = diff.perpendicular()
                d = -n.dot(self.points[i])
                self.leqs.append((n, d))
        return self.leqs

    def get_lines(self):  # get edge lines
        return [
            Line(self.points[i - 1], self.points[i])
            for i in range(len(self.points))
        ]

    def get_radius(self):  # the radius of polygon
        if self.radius is None:
            self.radius = max([point.length() for point in self.points])
        return self.radius

    def cspace_poly(self, other):  # the translational Cspace polygon
        # This is the Star algorithm
        poly = other.invert()
        my_angles = [vector[0].angle() for vector in self.get_leqs()]
        my_edges = [(my_angles[i], self.points[i - 1], self.points[i])
                    for i in range(len(self.points))]

        poly_angles = [vector[0].angle() for vector in poly.get_leqs()]
        poly_edges = [(poly_angles[i], poly.points[i - 1], poly.points[i])
                      for i in range(len(poly.points))]

        start_point = my_edges[0][1] + poly_edges[0][1]
        new_edges = merge(my_edges, poly_edges)

        verts = [start_point]
        for i in range(len(new_edges) - 1):
            (angle, p1, p2) = new_edges[i]
            verts.append(verts[-1] + (p2 - p1))

        return Polygon(verts)

    def invert(self):  # reflect the polygon
        # Used for Minkowski difference
        return Polygon([-1 * point for point in self.points])

    def translate(self, reference):  # translate polygon
        return Polygon([reference + point for point in self.points])

    def rotate(self, angle):  # rotate polygon about origin (counter-clockwise)
        return Polygon([point.rotate(angle) for point in self.points])

    def at_pose(self, pose):  # place polygon at pose
        return self.rotate(pose.orientation).translate(pose.position)

    def collides(self, other, edges=False):  # test collision with polygon
        # edges = True counts a point being on an edge as collision
        if not self.get_aabb().collides(other.get_aabb()):
            # bounding boxes do not overlap
            return False
        if isinstance(other, Polygon):  # polygon-polygobn collision
            if self.contains(other.points[0]) or other.contains(self.points[0]):
                return True
            for i in range(len(self.points)):
                for j in range(len(other.points)):
                    this_line = Line(self.points[i - 1], self.points[i])
                    other_line = Line(other.points[j - 1], other.points[j])
                    if this_line.intersect(other_line, edges=edges):
                        return True
            return False
        elif isinstance(other, Line):  # line-polygon collision
            if self.contains(other.midpoint()) or \
                    self.contains(other.start) or \
                    self.contains(other.end):
                return True
            for i in range(len(self.points)):
                if Line(self.points[i - 1],
                        self.points[i]).intersect(other, edges=edges):
                    return True
            return False
        else:
            raise Exception

    def contains(self, other, edges=False):  # does polygon contain other
        # edges = True counts a point being on an edge as containment
        if isinstance(other, Point):  # point in polygon
            if edges:
                return self.get_aabb().contains(other) and \
                    all([line[0].dot(other) + line[1]
                         <= 0 for line in self.get_leqs()])
            else:
                return self.get_aabb().contains(other) and \
                    all([line[0].dot(other) + line[1]
                         < 0 for line in self.get_leqs()])
        elif isinstance(other, Polygon):  # polygon in polygon
            return self.get_aabb().contains(other.get_aabb()) and \
                all([self.contains(vert, edges) for vert in other.points])
        elif isinstance(other, Line):  # Definition of convex object
            return self.contains(line.start, edges=edges) and self.contains(
                line.end, edges=edges)
        else:
            raise Exception

    def sweep_move(self,
                   direction):  # sweep convex polygon along direction (point)
        return [
            Polygon([
                self.points[i - 1], self.points[i], self.points[i] + direction,
                self.points[i - 1] + direction
            ]) for i in range(len(self.points))
        ]

    def draw(self, window, color='blue'):
        window.drawPolygon(self.points, color=color)

    def __eq__(self, other):
        return (isinstance(other, Polygon)
                and len(self.points) == len(other.points)
                and all(self.points[i] == other.points[i]
                for i in range(len(self.points))))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return 'Polygon: (' + str(self.points) + ')'

    def __hash__(self):
        return str(self).__hash__()

    __str__ = __repr__


# Represents the union of a set of convex polygons, the polygons are
# expressed relative to the reference point.
class Object:

    def __init__(self, reference, polys):
        self.reference = reference
        self.local_polys = polys  # relative to reference
        self.polys = [poly.translate(reference)
                      for poly in polys]  # in global coords
        self.aabb = None
        self.radius = None

    def get_aabb(self):  # bounding box
        if self.aabb is None:
            min_x = min([poly.get_aabb().min.x for poly in self.polys])
            min_y = min([poly.get_aabb().min.y for poly in self.polys])
            max_x = max([poly.get_aabb().max.x for poly in self.polys])
            max_y = max([poly.get_aabb().max.y for poly in self.polys])
            self.aabb = AABB(Point(min_x, min_y), Point(max_x, max_y))
        return self.aabb

    def get_lines(self):  # edge lines
        lines = []
        for poly in self.polys:
            lines += poly.get_lines()
        return lines

    def get_radius(self):  # radius for the whole object
        if self.radius is None:
            self.radius = max([poly.get_radius() for poly in self.local_polys])
        return self.radius

    def collides(self, other, edges=False):  # collision test
        # edges = True counts a point being on an edge as collision
        return self.get_aabb().collides(other.get_aabb(), edges) and \
            any([my_poly.collides(other_poly, edges)
                 for my_poly in self.polys for other_poly in other.polys])

    def contains(self, point, edges=False):  # containment test
        # edges = True counts a point being on an edge as collision
        return self.get_aabb().contains(point, edges) and \
            any([poly.contains(point, edges) for poly in self.polys])

    def rotate(self, angle):  # rotate about the reference (not local_polys)
        return Object(self.reference,
                      [poly.rotate(angle) for poly in self.local_polys])

    def translate(self, reference):  # translate to new reference
        return Object(reference, self.local_polys)

    def at_pose(self, pose):  # place at new pose
        return self.translate(pose.position).rotate(pose.orientation)

    def sweep_move(self, start, finish):  # swept volume
        return Object(start, [
            swept for poly in self.local_polys
            for swept in poly.sweep_move(finish - start)
        ])

    def cspace_object(self, other):  # translationcal cspace object
        new_polys = []
        for my_poly in self.local_polys:
            for other_poly in other.local_polys:
                new_polys.append(my_poly.cspace_poly(other_poly))
        return Object(self.reference, new_polys)

    def draw(self, window, color='blue'):
        for poly in self.polys:
            poly.draw(window, color)
        # self.get_aabb().draw(window, color=color, width = 3)
        self.reference.draw(window, color='black', radius=3)

    def __repr__(self):
        return 'Object: (' + str(self.reference) + ", " + str(self.polys) + ')'

    def __hash__(self):
        return str(self).__hash__()

    __str__ = __repr__


# Convex hull polygon given set of points.
def convex_hull(verts):
    points = [(v.x, v.y) for v in verts]
    return [Point(v[0], v[1]) for v in convexHull(points)[::-1]]


######################################################################
# Convex hull code from http://python.net/~gherman/convexhull.html
######################################################################


def _myDet(p, q, r):
    # We use Sarrus' Rule to calculate the determinant.
    # (could also use the Numeric package...)
    sum1 = q[0] * r[1] + p[0] * q[1] + r[0] * p[1]
    sum2 = q[0] * p[1] + r[0] * q[1] + p[0] * r[1]

    return sum1 - sum2


def _isRightTurn(p, q, r):
    assert p != q and q != r and p != r

    if _myDet(p, q, r) < 0:
        return 1
    else:
        return 0


def convexHull(P):
    # Get a local list copy of the points and sort them lexically.
    points = sorted(map(None, P))

    # Build upper half of the hull.
    upper = [points[0], points[1]]
    for p in points[2:]:
        upper.append(p)
        while len(upper) > 2 and not _isRightTurn(upper[-3:]):
            del upper[-2]

    # Build lower half of the hull.
    points.reverse()
    lower = [points[0], points[1]]
    for p in points[2:]:
        lower.append(p)
        while len(lower) > 2 and not _isRightTurn(lower[-3:]):
            del lower[-2]

    # Remove duplicates.
    del lower[0]
    del lower[-1]

    # Concatenate both halfs and return.
    return tuple(upper + lower)



import typing
import unittest
from functools import cache

from pydrake.all import (
    AddFrame,
    AddModel,
    AddWeld,
    IiwaDriver,
    ModelDirective,
    MultibodyPlant,
    RobotDiagram,
    Rotation,
    Transform,
)

from manipulation.directives_tree import DirectivesTree
from manipulation.station import (
    InverseDynamicsDriver,
    MakeHardwareStation,
    MakeMultibodyPlant,
    Scenario,
)


class DirectivesTreeTest(unittest.TestCase):
    @cache
    def get_flattened_directives(
        self, mobile_iiwa: bool = False
    ) -> typing.List[ModelDirective]:
        if mobile_iiwa:
            # Add mobile iiwa.
            directives = [
                ModelDirective(
                    add_model=AddModel(
                        name="iiwa",
                        file="package://manipulation/mobile_iiwa14_primitive_collision.urdf",
                    )
                ),
            ]
        else:
            # Add iiwa and weld it to the world frame.
            directives = [
                ModelDirective(
                    add_model=AddModel(
                        name="iiwa",
                        file="package://drake_models/iiwa_description/urdf/iiwa14_no_collision.urdf",
                    )
                ),
                ModelDirective(
                    add_frame=AddFrame(
                        name="iiwa_origin",
                        X_PF=Transform(
                            base_frame="world",
                            translation=[0, 0.765, 0],
                        ),
                    )
                ),
                ModelDirective(
                    add_weld=AddWeld(
                        parent="iiwa_origin",
                        child="iiwa::base",
                    )
                ),
            ]
        # Add and weld wsg + table.
        directives += [
            ModelDirective(
                add_model=AddModel(
                    name="wsg",
                    file="package://drake_models/wsg_50_description/sdf/schunk_wsg_50_with_tip.sdf",
                )
            ),
            ModelDirective(
                add_frame=AddFrame(
                    name="iiwa::wsg_attach",
                    X_PF=Transform(
                        base_frame="iiwa::iiwa_link_7",
                        translation=[0, 0, 0.114],
                        rotation=Rotation.Rpy(deg=[90.0, 0.0, 68.0]),
                    ),
                )
            ),
            ModelDirective(
                add_weld=AddWeld(
                    parent="iiwa::wsg_attach",
                    child="wsg::body",
                )
            ),
            ModelDirective(
                add_model=AddModel(
                    name="table",
                    file="package://drake_models/manipulation_station/table_wide.sdf",
                )
            ),
            ModelDirective(
                add_frame=AddFrame(
                    name="table_origin",
                    X_PF=Transform(
                        base_frame="world",
                        translation=[0.4, 0.3825, 0.0],
                        rotation=Rotation.Rpy(deg=[0.0, 0.0, 0.0]),
                    ),
                )
            ),
            ModelDirective(
                add_weld=AddWeld(
                    parent="table_origin",
                    child="table::table_body",
                )
            ),
        ]
        return directives

    def test_get_welded_descendants_and_directives(self):
        directives = self.get_flattened_directives()
        tree = DirectivesTree(directives)

        children, wsg_directives = tree.GetWeldedDescendantsAndDirectives(["iiwa"])
        self.assertEqual(children, {"wsg"})
        sorted_wsg_directives = tree.TopologicallySortDirectives(wsg_directives)
        self.assertEqual(
            sorted_wsg_directives, directives[3:6]
        )  # wsg-related directives

    def test_get_weld_to_world_directives(self):
        directives = self.get_flattened_directives()
        tree = DirectivesTree(directives)

        iiwa_directives = tree.GetDirectivesFromModelsToRoot(["iiwa"])
        sorted_iiwa_directives = tree.TopologicallySortDirectives(iiwa_directives)
        self.assertEqual(
            sorted_iiwa_directives, directives[:3]
        )  # iiwa-related directives

        iiwa_wsg_directives = tree.GetDirectivesFromModelsToRoot(["iiwa", "wsg"])
        sorted_iiwa_wsg_directives = tree.TopologicallySortDirectives(
            iiwa_wsg_directives
        )
        self.assertEqual(sorted_iiwa_wsg_directives, directives[:6])

    def make_multibody_plant_tester(self, scenario: Scenario, mobile_iiwa: bool):
        num_wsg_positions = 2
        num_iiwa_positions = 10 if mobile_iiwa else 7

        # Test with all model instances.
        plant: MultibodyPlant = MakeMultibodyPlant(scenario)
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(plant.HasModelInstanceNamed("wsg"))
        self.assertTrue(plant.HasModelInstanceNamed("table"))
        self.assertEqual(plant.num_positions(), num_iiwa_positions + num_wsg_positions)

        # Test with only "iiwa" and "wsg".
        plant: MultibodyPlant = MakeMultibodyPlant(
            scenario, model_instance_names=["iiwa", "wsg"]
        )
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(plant.HasModelInstanceNamed("table"))
        self.assertEqual(plant.num_positions(), num_iiwa_positions + num_wsg_positions)

        # Test with only "iiwa" and "wsg" (reversed order).
        plant: MultibodyPlant = MakeMultibodyPlant(
            scenario, model_instance_names=["wsg", "iiwa"]
        )
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(plant.HasModelInstanceNamed("table"))
        self.assertEqual(plant.num_positions(), num_iiwa_positions + num_wsg_positions)

        # Test with only "iiwa".
        plant: MultibodyPlant = MakeMultibodyPlant(
            scenario, model_instance_names=["iiwa"]
        )
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertFalse(plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(plant.HasModelInstanceNamed("table"))
        self.assertEqual(plant.num_positions(), num_iiwa_positions)

        # Test with only "iiwa" and "add_frozen_child_instances=True"
        plant: MultibodyPlant = MakeMultibodyPlant(
            scenario,
            model_instance_names=["iiwa"],
            add_frozen_child_instances=True,
        )
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(plant.HasModelInstanceNamed("table"))
        self.assertEqual(plant.num_positions(), num_iiwa_positions)

    def test_make_multibody_plant_welded_iiwa(self):
        scenario = Scenario()
        scenario.directives = self.get_flattened_directives()
        scenario.model_drivers = {
            "iiwa": IiwaDriver(
                control_mode="position_only",
                hand_model_name="wsg",
            )
        }

        self.make_multibody_plant_tester(scenario, mobile_iiwa=False)

    def test_make_multibody_plant_mobile_iiwa(self):
        scenario = Scenario()
        scenario.directives = self.get_flattened_directives(mobile_iiwa=True)
        scenario.model_drivers = {"iiwa": InverseDynamicsDriver()}

        self.make_multibody_plant_tester(scenario, mobile_iiwa=True)

    def test_load_scenario_iiwa_driver(self):
        scenario = Scenario()
        scenario.directives = self.get_flattened_directives()
        scenario.model_drivers = {
            "iiwa": IiwaDriver(
                control_mode="position_only",
                hand_model_name="wsg",
            )
        }
        station: RobotDiagram = MakeHardwareStation(scenario, hardware=False)
        controller_plant: MultibodyPlant = station.GetSubsystemByName(
            "iiwa_controller_plant_pointer_system"
        ).get()

        # Check that the controller plant contains "iiwa" and "wsg" but not "table".
        self.assertTrue(controller_plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(controller_plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(controller_plant.HasModelInstanceNamed("table"))
        # Iiwa only driver => 7 positions.
        self.assertEqual(controller_plant.num_positions(), 7)

    def test_load_scenario_inverse_dynamics_driver(self):
        scenario = Scenario()
        scenario.directives = self.get_flattened_directives(mobile_iiwa=True)
        scenario.model_drivers = {"iiwa+wsg": InverseDynamicsDriver()}
        station: RobotDiagram = MakeHardwareStation(scenario, hardware=False)
        controller = station.GetSubsystemByName("iiwa+wsg.controller")
        controller_plant = controller.get_multibody_plant_for_control()

        # Check that the controller plant contains "iiwa" and "wsg" but not "table".
        self.assertTrue(controller_plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(controller_plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(controller_plant.HasModelInstanceNamed("table"))
        # Iiwa and wsg driver => 10+2=12 positions.
        self.assertEqual(controller_plant.num_positions(), 12)


if __name__ == "__main__":
    unittest.main()



import os
from datetime import date
from pathlib import Path
from urllib.request import urlretrieve
from warnings import warn

import numpy as np
import pydot
from IPython import get_ipython
from IPython.display import SVG, display
from pydrake.all import GetDrakePath
from pydrake.common import GetDrakePath
from pydrake.geometry import RenderLabel
from pydrake.multibody.parsing import PackageMap, Parser
from pydrake.systems.framework import System
from pydrake.systems.sensors import ImageLabel16I

# Use a global variable here because some calls to IPython will actually cause
# an interpreter to be created.  This file needs to be imported BEFORE that
# happens.
running_as_notebook = (
    "COLAB_TESTING" not in os.environ
    and get_ipython()
    and hasattr(get_ipython(), "kernel")
)

running_as_test = False


def _set_running_as_test(value: bool):
    """[INTERNAL USE ONLY]: Set the global variable `running_as_test` to
    `value`.

    This method is used by the build system; it is not intended for general
    use.
    """
    global running_as_test
    running_as_test = value


def FindResource(filename: str):
    """Returns the absolute path to the given filename relative to the
    manipulation module."""
    return os.path.join(os.path.dirname(__file__), filename)


def LoadDataResource(filename: str):
    warn("`LoadDataResource` is deprecated. Use `FindDataResource` instead.")
    FindDataResource(filename)


def FindDataResource(filename: str):
    """
    Returns the absolute path to the given filename relative to the book data
    directory; fetching it from a remote host if necessary.
    """
    if "MANIPULATION_DATA_DIR" in os.environ:
        data = os.environ["MANIPULATION_DATA_DIR"]
    else:
        data = os.path.join(os.path.dirname(os.path.dirname(__file__)), "book/data")
        if not os.path.exists(data):
            os.makedirs(data)
    path = os.path.join(data, filename)
    if not os.path.exists(path):
        if running_as_test:
            raise FileNotFoundError(
                f"{path} was not found locally; it is required for testing."
            )  # because the urlretrieve defeats bazel's hermetic testing.
        print(f"{path} was not found locally; downloading it now...")
        urlretrieve(f"https://manipulation.csail.mit.edu/data/{filename}", path)
    return path


def ConfigurePackageMap(package_map: PackageMap):
    """Add the `manipulation` module packages to the given PackageMap."""
    package_xml = os.path.join(os.path.dirname(__file__), "models/package.xml")
    package_map.AddPackageXml(filename=package_xml)


def ConfigureParser(parser: Parser):
    """Add the `manipulation` module packages to the given Parser."""
    ConfigurePackageMap(parser.package_map())


def AddMujocoMenagerie(package_map: PackageMap):
    """Add the remote `mujoco_menagerie` package to the given PackageMap.
    https://github.com/google-deepmind/mujoco_menagerie"""
    package_map.AddRemote(
        package_name="mujoco_menagerie",
        params=PackageMap.RemoteParams(
            # This repository doesn't have up-to-date tags/releases; the scary
            # hash in the url is the most recent commit sha at the time of my
            # writing.
            urls=[
                f"https://github.com/google-deepmind/mujoco_menagerie/archive/469893211c41d5da9c314f5ab58059fa17c8e360.tar.gz"
            ],
            sha256=("1cfe0ebde2c6dd80405977e0b3a6f72e1b062d8a79f9f0437ebebe463c9c85f7"),
            strip_prefix="mujoco_menagerie-469893211c41d5da9c314f5ab58059fa17c8e360/",
        ),
    )


def colorize_labels(image: ImageLabel16I):
    """Given a label image, replace the integer labels with color values that display nicely in matplotlib."""

    # import needs to happen after backend is set up.
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    reserved_labels = [
        RenderLabel.kDoNotRender,
        RenderLabel.kDontCare,
        RenderLabel.kEmpty,
        RenderLabel.kUnspecified,
    ]

    """Colorizes labels."""
    # TODO(eric.cousineau): Revive and use Kuni's palette.
    cc = mpl.colors.ColorConverter()
    color_cycle = plt.rcParams["axes.prop_cycle"]
    colors = np.array([cc.to_rgb(c["color"]) for c in color_cycle])
    bg_color = [0, 0, 0]
    image = np.squeeze(image)
    background = np.zeros(image.shape[:2], dtype=bool)
    for label in reserved_labels:
        background |= image == int(label)
    image[np.logical_not(background)]
    color_image = colors[image % len(colors)]
    color_image[background] = bg_color
    return color_image


def GetDrakeVersion() -> str:
    """Returns a string representing the Drake version, if availableor "source"."""
    drake_version_txt = (
        Path(GetDrakePath()).parent.parent / "doc" / "drake" / "VERSION.TXT"
    )
    if drake_version_txt.is_file():
        with open(drake_version_txt, "r") as f:
            drake_version = f.read()
        drake_version = drake_version.split()[0]
    else:
        drake_version = "source"
    return drake_version


def DrakeVersionGreaterThan(minimum_date: date):
    """Check that the Drake version is at least `minimum_data`."""
    drake_version_txt = (
        Path(GetDrakePath()).parent.parent / "doc" / "drake" / "VERSION.TXT"
    )
    version_dates = {
        "1.13.0": date(year=2023, month=2, day=14),
        "1.14.0": date(year=2023, month=3, day=15),
        "1.15.0": date(year=2023, month=4, day=18),
        "1.16.0": date(year=2023, month=5, day=18),
        "1.17.0": date(year=2023, month=5, day=23),
        "1.18.0": date(year=2023, month=6, day=20),
        "1.19.0": date(year=2023, month=7, day=13),
        "1.20.0": date(year=2023, month=8, day=16),
        "1.21.0": date(year=2023, month=9, day=14),
        "1.22.0": date(year=2023, month=10, day=16),
        "1.23.0": date(year=2023, month=11, day=17),
        "1.24.0": date(year=2023, month=12, day=18),
    }
    # If the file doesn't exist, then we should pass. A source install won't
    # have VERSION.TXT
    if drake_version_txt.is_file():
        with open(drake_version_txt, "r") as f:
            drake_version = f.read()
        drake_version = drake_version.split()[0]
        if len(drake_version) == 14:
            drake_date = date(
                year=int(drake_version[:4]),
                month=int(drake_version[4:6]),
                day=int(drake_version[6:8]),
            )
        elif drake_version in version_dates:
            drake_date = version_dates[drake_version]
        else:
            warn(f"Unrecognized drake version {drake_version}")
            return
        if drake_date < minimum_date:
            raise (
                RuntimeError(
                    f"You need to update your Drake version. Python is using the Drake installation in {GetDrakePath()}. This installation was from a nightly build on {drake_date}, but this method requires Drake from at least {minimum_date}."
                )
            )


def RenderDiagram(system: System, max_depth: int | None = None):
    """Use pydot to render the GraphViz diagram of the given system.

    Args:
        system (System): The Drake system (or diagram) to render.
        max_depth (int, optional): Sets a limit to the depth of nested diagrams
            to visualize. Use zero to render a diagram as a single system
            block. Defaults to 1.
    """
    display(
        SVG(
            pydot.graph_from_dot_data(system.GetGraphvizString(max_depth=max_depth))[
                0
            ].create_svg()
        )
    )


# Adapted from Drake's system_doxygen.py.  Just make an html rendering of the
# system block with its name and input/output ports (even if it is a Diagram).
def SystemHtml(system: System):
    """Generates an HTML string for `system` of the style seen in the Drake
    doxygen documentation."""
    input_port_html = ""
    for p in range(system.num_input_ports()):
        input_port_html += (
            f'<tr><td align=right style="padding:5px 0px 5px 0px">'
            f"{system.get_input_port(p).get_name()} &rarr;</td></tr>"
        )
    output_port_html = ""
    for p in range(system.num_output_ports()):
        output_port_html += (
            '<tr><td align=left style="padding:5px 0px 5px 0px">'
            f"&rarr; {system.get_output_port(p).get_name()}</td></tr>"
        )
    # Note: keeping this on a single line avoids having to handle comment line
    # markers (e.g. * or ///)
    html = (
        f"<table align=center cellpadding=0 cellspacing=0><tr align=center>"
        f'<td style="vertical-align:middle">'
        f"<table cellspacing=0 cellpadding=0>{input_port_html}</table>"
        f"</td>"
        f'<td align=center style="border:2px solid black;padding-left:20px;'
        f'padding-right:20px;vertical-align:middle" bgcolor=#F0F0F0>'
        f"{system.get_name()}</td>"
        f'<td style="vertical-align:middle">'
        f"<table cellspacing=0 cellpadding=0>{output_port_html}</table>"
        f"</td></tr></table>"
    )

    return html
